/* 
 * Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of the
 * License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301  USA
 */


#include "stdafx.h"

#include "wb_sql_editor_form.h"
#include "sqlide/recordset_be.h"
#include "sqlide/recordset_cdbc_storage.h"
#include "grtsqlparser/sql_facade.h"
#include "grtui/grtdb_connection_editor.h"
#include "grtui/eol_format_dialog.h"
#include "grtui/confirm_save_dialog.h"
#include "sqlide/wb_sql_editor_snippets.h"

#include "objimpl/db.query/db_query_Resultset.h"

#include "workbench/wb_context_ui.h"

#include "grtdb/editor_dbobject.h"
#include "grtdb/editor_table.h"

#include "diff/diffchange.h"
#include "grtdb/diff_dbobjectmatch.h"
#include "db.mysql/src/module_db_mysql.h"

#include "sqlide/sql_script_run_wizard.h"
#include "grtui/file_charset_dialog.h"
#include "sqlide/recordset_export_form.h"

#include "base/file_functions.h"
#include "base/util_functions.h"

// this doesn't belong here, but there's no other workaround for having access to mysql_info
// because of that we also need to link wbprivate directly to the connector, making their
// dynamic loading pretty much useless
#include <cppconn/../driver/mysql_connection.h>

#include "base/boost_smart_ptr_helpers.h"

#include <errno.h>
#include <glib/gstdio.h>

#include <boost/foreach.hpp>
#include <boost/scoped_ptr.hpp>
#include <locale>
#include <algorithm>
#include <cctype>
#include <sstream>


using namespace bec;
using namespace grt;
using namespace wb;
using namespace base;

// 5M max file size for auto-restoring
#define MAX_FILE_SIZE_FOR_AUTO_RESTORE 5000000

// in the Actions log, truncate anything that's longer than this
#define MAX_LOG_STATEMENT_TEXT 4098


const char *SQL_EXCEPTION_MSG_FORMAT= _("Error Code: %i\n%s");
const char *EXCEPTION_MSG_FORMAT= _("Error: %s");

#define CATCH_SQL_EXCEPTION_AND_DISPATCH(statement, log_message_index, duration) \
catch (sql::SQLException &e)\
{\
  set_log_message(log_message_index, grt::ErrorMsg, strfmt(SQL_EXCEPTION_MSG_FORMAT, e.getErrorCode(), e.what()), statement, duration);\
}

#define CATCH_EXCEPTION_AND_DISPATCH(statement) \
catch (std::exception &e)\
{\
  add_log_message(grt::ErrorMsg, strfmt(EXCEPTION_MSG_FORMAT, e.what()), statement, "");\
}

#define CATCH_ANY_EXCEPTION_AND_DISPATCH(statement) \
catch (sql::SQLException &e)\
{\
  add_log_message(grt::ErrorMsg, strfmt(SQL_EXCEPTION_MSG_FORMAT, e.getErrorCode(), e.what()), statement, "");\
}\
CATCH_EXCEPTION_AND_DISPATCH(statement)

#define CATCH_ANY_EXCEPTION_AND_DISPATCH_TO_DEFAULT_LOG(statement) \
catch (sql::SQLException &e)\
{\
  _grtm->get_grt()->send_error(strfmt(SQL_EXCEPTION_MSG_FORMAT, e.getErrorCode(), e.what()), statement);\
}\
catch (std::exception &e)\
{\
  _grtm->get_grt()->send_error(strfmt(EXCEPTION_MSG_FORMAT, e.what()), statement);\
}


class Timer
{
public:
  Timer(bool run_immediately) : _is_running(false), _start_timestamp(0), _duration(0)
  {
    if (run_immediately)
      run();
  }
  void reset()
  {
    _is_running= false;
    _start_timestamp= 0;
    _duration= 0;
  }
  void run()
  {
    if (_is_running)
      return;
    _is_running= true;
    _start_timestamp= timestamp();
  }
  void stop()
  {
    if (!_is_running)
      return;
    _is_running= false;
    _duration+= timestamp() - _start_timestamp;
  }
  time_t duration()
  {
    return _is_running ? (_duration + timestamp() - _start_timestamp) : _duration;
  }
  std::string duration_formatted()
  {
    std::div_t s;
    s= std::div((int)duration(), 1000);
    return strfmt(_("%i.%.3i sec"), s.quot, s.rem);
  }
private:
  bool _is_running;
  time_t _start_timestamp;
  time_t _duration;
};


std::string & strip_sql_ident(std::string &text)
{
  std::locale loc;

  size_t size= text.size();

  size_t start= 0;
  for (; start < size; ++start)
    if (!std::isspace(text[start], loc) && (text[start] != '`') && (text[start] != '\'') && (text[start] != '"'))
      break;

  size_t end= size;
  for (; end > 0; --end)
    if (!std::isspace(text[end-1], loc) && (text[end-1] != '`') && (text[end-1] != '\'') && (text[end-1] != '"'))
      break;

  return text= text.substr(start, end-start);
}


Db_sql_editor::Ref Db_sql_editor::create(wb::WBContextSQLIDE *wbsql, const db_mgmt_ConnectionRef &conn)
{
  Db_sql_editor::Ref instance(new Db_sql_editor(wbsql, conn)); 
  if (conn.is_valid())
  {
    instance->connect();
    instance->finish_startup();
  }
  return instance;
}


Db_sql_editor::Db_sql_editor(wb::WBContextSQLIDE *wbsql, const db_mgmt_ConnectionRef &conn)
:
_wbsql(wbsql),
_context_ui(_wbsql->get_wbui()),
_grtm(wbsql->get_grt_manager()),
_menu(0),
_toolbar(0),
_loading_workspace(false),
_sql_editors_serial(0),
_scratch_editors_serial(0),
_active_sql_editor_index(0),
_updating_sql_editor(0),
_keep_alive_thread(NULL),
_connection(conn),
_aux_dbc_conn(new sql::Dbc_connection_handler()),
_usr_dbc_conn(new sql::Dbc_connection_handler()),
exec_sql_task(GrtThreadedTask::create(wbsql->get_grt_manager())),
_is_running_query(false),
_recordsets_are_pinned_by_default(false),
_rs_sequence(0),
_schema_tree(wbsql->get_grt_manager()->get_grt()),
_side_bar(0),
live_schema_fetch_task(GrtThreadedTask::create(wbsql->get_grt_manager())),
_side_splitter(false, true),
_info_box((mforms::ScrollBars)(mforms::BothScrollBars|mforms::SmallScrollBars)),
_live_physical_overview(NULL),
live_schemata_refresh_task(GrtThreadedTask::create(wbsql->get_grt_manager())),
_log(DbSqlEditorLog::create(wbsql->get_grt_manager())),
_history(DbSqlEditorHistory::create(wbsql->get_grt_manager()))
{  
  _options= DictRef::cast_from(_grtm->get_grt()->get("/wb/options/options"));

  _wbsql->get_wbui()->signal_main_form_change().connect(sigc::mem_fun(this, &Db_sql_editor::main_form_changed));
  
  _sql_editors_mutex= g_mutex_new();
  _aux_dbc_conn_mutex= g_mutex_new();
  _usr_dbc_conn_mutex= g_mutex_new();
  _keep_alive_thread_mutex= g_mutex_new();
  _schema_contents_mutex= g_mutex_new();
  _recordsets_mutex= g_mutex_new();

  _schema_tree.is_schema_contents_enabled(_options.get_int("DbSqlEditor:ShowSchemaTreeSchemaContents", 1) != 0);
  _schema_tree.refresh_signal.connect(sigc::bind(sigc::mem_fun(this, &Db_sql_editor::refresh_physical_overview), false));
  _schema_tree.active_schema_change_signal.connect(sigc::mem_fun(this, (void(Db_sql_editor::*)(const std::string&))&Db_sql_editor::active_schema));
  _schema_tree.list_fetch_slot= sigc::mem_fun(this, &Db_sql_editor::fetch_live_schema_list);
  _schema_tree.fetch_slot= sigc::mem_fun(this, &Db_sql_editor::fetch_live_schema_contents);
  _schema_tree.details_fetch_slot= sigc::mem_fun(this, &Db_sql_editor::fetch_object_details);
  _schema_tree.sql_editor_text_insert_signal.connect(sql_editor_text_insert_signal.make_slot());
  _schema_tree.activate_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::schema_object_activated));
  _schema_tree.alter_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_alter_live_object));
  _schema_tree.create_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_create_live_object));
  _schema_tree.drop_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_drop_live_object));

  _schema_tree.object_plugin_items_slot= sigc::mem_fun(this, &Db_sql_editor::get_live_catalog_menu_items);
  _schema_tree.call_object_plugin_items_slot= sigc::mem_fun(this, &Db_sql_editor::call_live_catalog_menu_item);

  _live_physical_overview= new LivePhysicalOverviewBE(_context_ui->get_wb());
  _is_physical_overview_enabled= (_options.get_int("DbSqlEditor:ShowLivePhysicalOverview", 1) != 0);
  if (_is_physical_overview_enabled)
  {
    _live_physical_overview->refresh_signal.connect(sigc::bind(sigc::mem_fun(this, &Db_sql_editor::refresh_physical_overview), false));
    _live_physical_overview->active_schema_change_signal.connect(sigc::mem_fun(this, (void(Db_sql_editor::*)(const std::string&))&Db_sql_editor::active_schema));
    //!_live_physical_overview->schema_list_fetch_slot= sigc::mem_fun(this, &Db_sql_editor::fetch_live_schema_list);
    _live_physical_overview->schema_fetch_slot= sigc::mem_fun(this, &Db_sql_editor::fetch_live_schema_contents);
    _live_physical_overview->activate_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::schema_object_activated));
    _live_physical_overview->alter_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_alter_live_object));
    //!_live_physical_overview->create_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_create_live_object));
    _live_physical_overview->drop_live_object_signal.connect(sigc::mem_fun(this, &Db_sql_editor::do_drop_live_object));
    _live_physical_overview->active_schema_index= sigc::mem_fun(this, &Db_sql_editor::active_schema_index);
    
    _live_physical_overview->object_plugin_items_slot= sigc::mem_fun(this, &Db_sql_editor::get_live_catalog_menu_items);
    _live_physical_overview->call_object_plugin_items_slot= sigc::mem_fun(this, &Db_sql_editor::call_live_catalog_menu_item);
  }

  live_schemata_refresh_task->send_task_res_msg(false);
  live_schemata_refresh_task->msg_cb(sigc::bind(sigc::mem_fun(this, &Db_sql_editor::add_log_message), ""));

  live_schema_fetch_task->send_task_res_msg(false);
  live_schema_fetch_task->msg_cb(sigc::bind(sigc::mem_fun(this, &Db_sql_editor::add_log_message), ""));

  exec_sql_task->send_task_res_msg(false);
  exec_sql_task->msg_cb(sigc::bind(sigc::mem_fun(this, &Db_sql_editor::add_log_message), ""));

  _last_log_message_timestamp= timestamp();

  _progress_status_update_interval= _options.get_int("DbSqlEditor:ProgressStatusUpdateInterval", 500);

  int keep_alive_interval= _options.get_int("DbSqlEditor:KeepAliveInterval", 600);
  if (keep_alive_interval != 0)
  {
    _keep_alive_thread= TimerActionThread::create(sigc::mem_fun(this, &Db_sql_editor::send_message_keep_alive), keep_alive_interval*1000);
    _keep_alive_thread->on_exit.connect(sigc::mem_fun(this, &Db_sql_editor::reset_keep_alive_thread));
  }

  _continue_on_error= (_options.get_int("DbSqlEditor:ContinueOnError", 0) != 0);

  register_default_actions();
}


void Db_sql_editor::finish_startup()
{
  setup_sidebar();

  if (!load_workspace(_connection->name()))
    add_sql_editor(true); // start with 1 text-buffer for editing sql
  
  // update the info box
  schema_row_selected();
}


Db_sql_editor::~Db_sql_editor()
{
  delete _side_bar;
  delete _toolbar;
  delete _menu;
  reset();

  reset_keep_alive_thread();

  delete _live_physical_overview;

  g_mutex_free(_recordsets_mutex);
  g_mutex_free(_schema_contents_mutex);
  g_mutex_free(_keep_alive_thread_mutex);
  g_mutex_free(_usr_dbc_conn_mutex);
  g_mutex_free(_aux_dbc_conn_mutex);
  g_mutex_free(_sql_editors_mutex);
}


void Db_sql_editor::auto_save()
{
  save_workspace(_connection->name(), true);
}


// Save all script buffers, including scratch buffers
void Db_sql_editor::save_workspace(const std::string &workspace_name, bool is_autosave)
{
  GMutexLock sql_editors_mutex(_sql_editors_mutex);

  if (_sql_editors.size() < 1 || (_sql_editors.size() == 1 && _sql_editors[0].is_scratch && _sql_editors[0].editor->sql().empty()))
    return;
  
  std::string path;
  
  // if we're autosaving, just use the same path from previous saves
  if (!is_autosave || _autosave_path.empty())
  {
    std::string path_prefix = bec::make_path(_grtm->get_user_datadir(),
                                             "sql_workspaces");
    if (!g_file_test(path_prefix.c_str(), G_FILE_TEST_EXISTS))
    {
      if (g_mkdir_with_parents(path_prefix.c_str(), 0700) < 0)
        throw std::runtime_error(base::strfmt("Could not create directory %s: %s", path_prefix.c_str(), g_strerror(errno)));
    }
    
    int i= 1;
    do
    {
      path = bec::make_path(path_prefix, base::strfmt("%s-%i%s", workspace_name.c_str(), i++, (is_autosave ? ".autosave" : ".workspace")));
    }
    while (g_mkdir(path.c_str(), 0700) < 0 && errno == EEXIST);
    
    if (is_autosave)
    {
      _autosave_path = path;
    }
  }
  else
    path = _autosave_path;

  // save the real id of the connection
  if (_connection.is_valid())
    g_file_set_contents(bec::make_path(path, "connection_id").c_str(), 
                        _connection->id().c_str(), _connection->id().size(), NULL);
  
  //XXX The editor contents need to be up-to-date. Since we can only request a refresh of the backend
  // for the active tab, the frontend must automatically refresh tabs when switching away from them
  // This can be removed once backend copy of text is dropped and getting data from frontend editor happens
  // through a callback
  
  int editor_index = 0;
  BOOST_FOREACH (Sql_editor_info sql_editor_info, _sql_editors)
  {
    editor_index++;
    
    // if the editor is saved to a file, we store the path. if the file is unsaved, we save an updated copy
    // for scrath areas, the file is always kept as an autosave
    std::string filename;
    
    if (!sql_editor_info.is_scratch)
    {
      GError *error = 0;
      std::string contents = sql_editor_info.filename + "\n" + sql_editor_info.orig_encoding;
      // save the path to the real file
      if (!g_file_set_contents(bec::make_path(path, base::strfmt("%i.filename", editor_index)).c_str(),
                               contents.data(), contents.size(), &error))
      {
        std::string msg(base::strfmt("Could not save snapshot of editor contents to %s: %s", path.c_str(), error->message));
        g_error_free(error);
        base_rmdir_recursively(path.c_str());
        throw msg;
      }
    }

    if (sql_editor_info.is_scratch)
      filename = base::strfmt("%i.scratch", editor_index);
    else
      if (sql_editor_info.dirty)
        filename = base::strfmt("%i.autosave", editor_index);
    
    // only save editor contents for scratch areas and unsaved editors
    if (!filename.empty())
    {
      const std::string &sql(sql_editor_info.editor->sql());
      GError *error = 0;
      std::string fn = bec::make_path(path, filename);
      
      if (!g_file_set_contents(fn.c_str(), sql.data(), sql.size(), &error))
      {
        std::string msg(base::strfmt("Could not save snapshot of editor contents to %s: %s", fn.c_str(), error->message));
        g_error_free(error);
        base_rmdir_recursively(path.c_str());
        throw msg;
      }
    }
  }
}


static bool check_if_file_too_big_to_restore(const std::string &path, const std::string &file_caption,
                                             bool allow_save_as = false)
{
  boost::int64_t length;
  if ((length = get_file_size(path.c_str())) > MAX_FILE_SIZE_FOR_AUTO_RESTORE)
  {
  again:
    std::string max_size = base::sizefmt(MAX_FILE_SIZE_FOR_AUTO_RESTORE, false);
    std::string size = base::sizefmt(length, false);
    int rc = mforms::Utilities::show_warning("Restore Workspace",
      base::strfmt("%s is bigger than %s (%s), would you like to restore it anyway?",
        file_caption.c_str(), max_size.c_str(), size.c_str()
      ),
      "Restore", "Skip", allow_save_as ? "Save As..." : ""
    );
    if (rc == mforms::ResultCancel)
      return false;
    
    if (rc == mforms::ResultOther)
    {
      mforms::FileChooser fchooser(mforms::SaveFile);
      
      fchooser.set_title(_("Save File As..."));
      if (fchooser.run_modal())
      {
        if (!copy_file(path.c_str(), fchooser.get_path().c_str()))
        {
          if (mforms::Utilities::show_error("Save File",
                                            base::strfmt("File %s could not be saved.", fchooser.get_path().c_str()),
                                            _("Retry"), _("Cancel"), "") == mforms::ResultOk)
            goto again;
        }
      }
      else
        goto again;
      return false;
    }
  }
  return true;
}


struct GuardBoolFlag
{
  bool *flag;
  GuardBoolFlag(bool *ptr) : flag(ptr) { *flag = true; }
  ~GuardBoolFlag() { *flag = false; }
};

// Restore a previously saved workspace for this connection. The loaded data is deleted immediately after loading (unless its an autosave)
bool Db_sql_editor::load_workspace(const std::string &workspace_name)
{
  std::string path_prefix = bec::make_path(_grtm->get_user_datadir(), "sql_workspaces");
  
  GuardBoolFlag flag(&_loading_workspace);

  // find workspaces on disk
  std::string workspace_path;
  bool restoring_autosave = false;
  {
    GDir *dir = g_dir_open(path_prefix.c_str(), 0, NULL);
    if (!dir)
      return false;
    int lowest_index = 9999999;
    const gchar *name;
    while ((name = g_dir_read_name(dir)) != NULL)
    {
      if (g_str_has_prefix(name, workspace_name.c_str()))
      {
        const char *end = strrchr(name, '.');
        int index = 0;
        if (end)
          index = atoi(std::string(name + workspace_name.size()+1, end-name-(workspace_name.size()+1)).c_str());
        
        if (g_str_has_suffix(name, ".autosave"))
        {
          if (!restoring_autosave)
          {
            lowest_index = index;
            restoring_autosave = true;
            workspace_path = name;
          }
          else
            if (index < lowest_index)
            {
              lowest_index = index;
              workspace_path = name;
            }
        }
        else if (!restoring_autosave && g_str_has_suffix(name, ".workspace"))
        {
          if (index < lowest_index)
          {
            workspace_path = name;
            lowest_index = index;
          }
        }
      }
    }
    g_dir_close(dir);
  }
  if (workspace_path.empty()) return false;
  workspace_path = bec::make_path(path_prefix, workspace_path);

  // get list of files
  std::vector<std::string> editor_files;
  {
    GDir *dir = g_dir_open(workspace_path.c_str(), 0, NULL);
    if (!dir)
      return false;
    const gchar *name;
    while ((name = g_dir_read_name(dir)) != NULL)
    {
      editor_files.push_back(name);
    }
    g_dir_close(dir);
  }
  std::sort(editor_files.begin(), editor_files.end());
  
  BOOST_FOREACH(std::string file, editor_files)
  {
    if (g_str_has_suffix(file.c_str(), ".scratch"))
    {
      GError *error = NULL;
      gchar *data;
      gsize length;
      
      if (!check_if_file_too_big_to_restore(bec::make_path(workspace_path, file),
                                           base::strfmt("Saved scratch buffer '%s'", file.c_str())))
        continue;
      
      if (!g_file_get_contents(bec::make_path(workspace_path, file).c_str(), &data, &length, &error))
      {
        int rc;
        rc = mforms::Utilities::show_error("Restore Workspace", 
                                           base::strfmt("Could not read contents of '%s'.\n%s", bec::make_path(workspace_path, file).c_str(), error->message),
                                           "Ignore", "Cancel", "");
        if (rc != mforms::ResultOk)
          return false;
      }
            
      int i = add_sql_editor(true);
      sql_editor_new_ui.emit(_active_sql_editor_index);
      _sql_editors[i].editor->sql(std::string(data, length));
      _sql_editors[i].dirty = true;
      
      _active_sql_editor_index = i;
      g_free(data);
    }
    else if (g_str_has_suffix(file.c_str(), ".filename"))
    {
      std::string autosave_filename = file.substr(0, file.size()-strlen(".filename")).append(".autosave");
      std::string filename;
      std::string encoding;
      GError *error = NULL;
      gchar *data;
      gsize length;
      
      if (!g_file_get_contents(bec::make_path(workspace_path, file).c_str(), &data, &length, NULL))
        filename = "";
      else
      {
        char *line1 = strtok(data, "\n");
        char *line2 = strtok(NULL, "\n");
        
        if (line1)
          filename = line1;
        if (line2)
          encoding = line2;
        g_free(data);
      }

      if (!g_file_test(bec::make_path(workspace_path, autosave_filename).c_str(), G_FILE_TEST_EXISTS))
      {
        
        if (check_if_file_too_big_to_restore(filename,
                                              base::strfmt("File '%s'", filename.c_str()),
                                              false))        
          // load the saved file
          open_sql_script_file(filename, false, false, encoding);
      }
      else
      {
        if (!check_if_file_too_big_to_restore(bec::make_path(workspace_path, autosave_filename).c_str(),
                                             base::strfmt("Auto-saved file '%s'", filename.c_str())))
          continue;

        // load auto-saved file and set the filename to the original path
        if (!g_file_get_contents(bec::make_path(workspace_path, autosave_filename).c_str(), &data, &length, &error))
        {
          int rc;
          rc = mforms::Utilities::show_error("Restore Workspace", 
                                             base::strfmt("Could not read auto-saved contents for '%s'.\n%s", 
                                                          (filename.empty() ? "Unsaved Script" : filename.c_str()), error->message),
                                             "Ignore", "Cancel", "");
          if (rc != mforms::ResultOk)
            return false;
        }
        else
        {
          int i = add_sql_editor(false);
          sql_editor_new_ui.emit(_active_sql_editor_index);
          _sql_editors[i].editor->sql(std::string(data, length));
          _sql_editors[i].filename = filename;
          _sql_editors[i].dirty = true;
          if (!filename.empty())
            _sql_editors[i].caption = g_basename(filename.c_str());
        
          _active_sql_editor_index = i;
          g_free(data);
        }
      }
    }
  }
  
  if (restoring_autosave)
  {
    _autosave_path = workspace_path;
    
    _grtm->replace_status_text(_("Restored last session state"));
  }
  else
  {
    base_rmdir_recursively(workspace_path.c_str());
  }
  
  return true;
}


void Db_sql_editor::reset_keep_alive_thread()
{
  GMutexLock keep_alive_thread_lock(_keep_alive_thread_mutex);
  if (_keep_alive_thread)
  {
    _keep_alive_thread->stop(true);
    _keep_alive_thread= NULL;
  }
}

bool Db_sql_editor::close()
{
  bool res= bec::UIForm::close();
  if (res)
  {
    if (_context_ui->get_wb()->get_root()->options()->options().get_int("workbench:SaveSQLWorkspaceOnClose", 0))
    {
      _grtm->replace_status_text("Saving workspace state...");
      if (_autosave_path.empty())
        save_workspace(_connection->name(), false);
      else
      {
        g_rename(_autosave_path.c_str(), (_autosave_path.substr(0, _autosave_path.size()-strlen(".autosave")) + ".workspace").c_str());
      }
    }
    else
    {
      if (!_autosave_path.empty())
        base_rmdir_recursively(_autosave_path.c_str());
    }

    _grtm->replace_status_text("Closing SQL Editor...");
    wbsql()->editor_will_close(this);

    exec_sql_task->disconnect_callbacks();
    reset_keep_alive_thread();
    close_connection(_usr_dbc_conn);
    close_connection(_aux_dbc_conn);
    
    _grtm->replace_status_text("SQL Editor closed");
  }
  return res;
}


std::string Db_sql_editor::get_form_context_name() const
{
  return WB_CONTEXT_QUERY;
}


void Db_sql_editor::activate_command(const std::string &command)
{
  wbsql()->get_wbui()->get_command_ui()->activate_command(command);
}


mforms::ToolBar *Db_sql_editor::get_toolbar()
{
  if (!_toolbar)
  {
    _toolbar = wbsql()->get_wbui()->get_command_ui()->create_toolbar("data/dbquery_toolbar.xml", sigc::mem_fun(this, &Db_sql_editor::activate_command));
        
    update_menu_and_toolbar();
  }
  return _toolbar;
}


mforms::MenuBar *Db_sql_editor::get_menubar()
{
  if (!_menu)
  {
    _menu = _context_ui->get_command_ui()->create_menubar_for_context(WB_CONTEXT_QUERY);

    update_menu_and_toolbar();
    
    _menu->set_item_enabled("query.save_edits", false);
    _menu->set_item_enabled("query.discard_edits", false);
    _menu->set_item_enabled("query.export", false);
    
    _menu->set_item_checked("query.stopOnError", !continue_on_error());
  }
  return _menu;
}


::ActionList & Db_sql_editor::action_list()
{
  return _action_list;
}


void Db_sql_editor::register_default_actions()
{
  //_action_list.register_action("exec_sql",
  //  sigc:sigc::mem_fun(this, &Db_sql_editor::exec_sql));

  //_action_list.register_action("exec_current_sql_statement",
  //  sigc::mem_fun(this, &Db_sql_editor::exec_current_sql_statement));

  _action_list.register_action("save_active_recordset_edits",
    sigc::mem_fun(this, &Db_sql_editor::save_active_recordset_edits));

  _action_list.register_action("discard_active_recordset_edits",
    sigc::mem_fun(this, &Db_sql_editor::discard_active_recordset_edits));
}


bool Db_sql_editor::get_session_variable(sql::Dbc_connection_handler::Ref &dbc_conn, const std::string &name, std::string &value)
{
  if (dbc_conn && dbc_conn->ref.get())
  {
    SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms());
    Sql_specifics::Ref sql_specifics= sql_facade->sqlSpecifics();
    std::string query= sql_specifics->query_variable(name);
    if (query.empty())
      return false;
    boost::scoped_ptr<sql::Statement> statement(dbc_conn->ref->createStatement());
    boost::scoped_ptr<sql::ResultSet> rs(statement->executeQuery(query));
    if (rs->next())
    {
      value= rs->getString(2);
      return true;
    }
  }
  return false;
}


void Db_sql_editor::cache_sql_mode()
{
  std::string sql_mode;
  if (get_session_variable(_usr_dbc_conn, "sql_mode", sql_mode))
  {
    if (sql_mode != _sql_mode)
    {
      _sql_mode= sql_mode;
      GMutexLock sql_editors_mutex(_sql_editors_mutex);
      BOOST_FOREACH (Sql_editor_info sql_editor_info, _sql_editors)
        sql_editor_info.editor->sql_mode(sql_mode);
    }
  }
}


int Db_sql_editor::add_sql_editor(bool scratch)
{
  Sql_editor::Ref sql_editor= Sql_editor::create(rdbms());
  sql_editor->eol(EolHelpers::eol());
  sql_editor->sql_check_progress_msg_throttle(_progress_status_update_interval);
  sql_editor->text_change_signal.connect(sigc::mem_fun(this, &Db_sql_editor::on_sql_editor_text_change));
  //sql_editor->text_selection_change_signal.connect(sigc::mem_fun(this, &Db_sql_editor::on_sql_editor_text_selection_change));
  sql_editor->sql_mode(_sql_mode);

  int sql_editor_index;
  {
    GMutexLock sql_editors_mutex(_sql_editors_mutex);
    Sql_editor_info info;
    info.filename = "";
    info.editor = sql_editor;
    info.dirty = false;
    info.is_scratch = scratch;
    if (!scratch)
      info.caption = strfmt("SQL File %i", ++_sql_editors_serial);
    else
      info.caption = strfmt("Query %i", ++_scratch_editors_serial);

    _sql_editors.push_back(info);
    sql_editor_index= _sql_editors.size() - 1;
    
    // create an empty initial resultset with a dummy data storage
    add_empty_recordset_for_sql_editor(sql_editor_index);
  }

  sql_editor_list_changed.emit(sql_editor, true);
  
  if (!_loading_workspace)
    auto_save();
  
  return sql_editor_index;
}


void Db_sql_editor::add_empty_recordset_for_sql_editor(int editor)
{
  Sql_editor_info &info(_sql_editors[editor]);
  {
    Recordset::Ref rs= Recordset::create(_grtm);
    rs->on_close.connect(sigc::mem_fun(*this, &Db_sql_editor::on_close_recordset));
    rs->caption(strfmt("%s Result", info.caption.c_str()));
    
    Recordset_cdbc_storage::Ref data_storage= Recordset_cdbc_storage::create(_grtm);
    data_storage->rdbms(rdbms());
    data_storage->dbms_conn(_usr_dbc_conn);    
    data_storage->reloadable(false);
    rs->data_storage(data_storage);
    
    GMutexLock lock(_recordsets_mutex);
    Recordsets::iterator iter = _recordsets.begin();
    while (iter != _recordsets.end())
    {
      int index = sql_editor_index_for_recordset((*iter)->key());
      if (index > editor)
      {
        _recordsets.insert(iter, rs);
        break;
      }
      ++iter;
    }
    if (iter == _recordsets.end())
      _recordsets.push_back(rs);
    info.resultsets.push_back(rs->key());
    recordset_list_changed.emit(rs, true);
  }  
}


void Db_sql_editor::remove_sql_editor(int index)
{
  {
    GMutexLock lock(_recordsets_mutex);
    // close unpinned recordsets attached to the editor
    Sql_editor_info &info(_sql_editors.at(index));
    for (std::vector<long long>::iterator r = info.resultsets.begin();
         r != info.resultsets.end(); ++r)
    {
      for (Recordsets::iterator rs = _recordsets.begin(); rs != _recordsets.end(); ++rs)
      {
        if ((*rs)->key() == *r)
        {
          if ((*rs)->pinned() || !(*rs)->can_close(false))
          {
            // leave it alone
          }
          else 
          {
            //delme close_recordset_ui.emit((*rs)->key());
            recordset_list_changed.emit(*rs, false);
            // close it
            _recordsets.erase(rs);
          }
          break;
        }
      }
    }
  }
  
  Sql_editor::Ref sql_editor;
  {
    GMutexLock sql_editors_mutex(_sql_editors_mutex);
    if ((index < 0) || (index >= (int)_sql_editors.size()))
      return;
    sql_editor= _sql_editors[index].editor;
    _sql_editors.erase(_sql_editors.begin() + index);
  }
  
  if (_active_sql_editor_index >= (int) _sql_editors.size())
    _active_sql_editor_index= _sql_editors.size()-1;
  
  sql_editor_list_changed.emit(sql_editor, false);
}


int Db_sql_editor::sql_editor_count()
{
  GMutexLock sql_editors_mutex(_sql_editors_mutex);
  return _sql_editors.size();
}


Sql_editor::Ref Db_sql_editor::sql_editor()
{
  if (_active_sql_editor_index < 0 || _active_sql_editor_index >= (int) _sql_editors.size())
    return Sql_editor::Ref();
  return _sql_editors.at(_active_sql_editor_index).editor;
}


Sql_editor::Ref Db_sql_editor::sql_editor(int index)
{
  GMutexLock sql_editors_mutex(_sql_editors_mutex);
  return _sql_editors.at(index).editor;
}


void Db_sql_editor::sql_editor_dirty(int index, bool flag)
{
  if (flag != _sql_editors.at(index).dirty)
  {
    _sql_editors.at(index).dirty = flag;

    do_partial_ui_refresh(RefreshEditorTitle);
  }
}

bool Db_sql_editor::sql_editor_will_close(int index)
{
  if (sql_editor_dirty(index))
  {
    int r;
    r = mforms::Utilities::show_warning(_("Close SQL Tab"), 
      base::strfmt(_("SQL script %s has unsaved changes.\n"
      "Would you like to Save these changes or Cancel closing?"),
      sql_editor_path(index).c_str()), _("Save"), _("Cancel"), _("Don't Save"));
    
    if (r == mforms::ResultCancel)
      return false;
    else if (r == mforms::ResultOk)
      if (!save_sql_script_file(sql_editor_path(index), index))
        return false;
  }
    
  return true;
}

std::string Db_sql_editor::sql_editor_caption(int index)
{
  if (index < 0)
    index = _active_sql_editor_index;
    
  if (index >= 0 && index < (int) _sql_editors.size())
  {
    if (_sql_editors.at(index).dirty && !_sql_editors.at(index).is_scratch)
      return _sql_editors.at(index).caption+"*";
    else
      return _sql_editors.at(index).caption;
  }
  return "";
}


long long Db_sql_editor::active_recordset_for_sql_editor(int index)
{
  Sql_editor_info &info(_sql_editors[index]);
  if (info.resultsets.empty())
    return -1;
  if (std::find(info.resultsets.begin(), info.resultsets.end(), info.active_resultset) != info.resultsets.end())
    return info.active_resultset;
  return info.resultsets.front();
}


int Db_sql_editor::sql_editor_index_for_recordset(long long rset)
{
  int i = 0;
  for (Sql_editors::iterator iter = _sql_editors.begin(); iter != _sql_editors.end(); ++iter)
  {
    if (std::find(iter->resultsets.begin(), iter->resultsets.end(), rset) != iter->resultsets.end())
      return i;
    ++i;
  }
  return -1;
}


void Db_sql_editor::reset()
{
  //_log->reset();
  _rs_sequence= 0;
}


RowId Db_sql_editor::add_log_message(int msg_type, const std::string &msg, const std::string &context, const std::string &duration)
{
  RowId new_log_message_index= _log->add_message(msg_type, truncate_text(context, MAX_LOG_STATEMENT_TEXT), msg, duration);
  _has_pending_log_messages= true;
  refresh_log_messages(false);
  if (msg_type == grt::ErrorMsg || msg_type == grt::WarningMsg)
    _exec_sql_error_count++;
  return new_log_message_index;
}


void Db_sql_editor::set_log_message(RowId log_message_index, int msg_type, const std::string &msg, const std::string &context, const std::string &duration)
{
  _log->set_message(log_message_index, msg_type, truncate_text(context, MAX_LOG_STATEMENT_TEXT), msg, duration);
  _has_pending_log_messages= true;
  if (msg_type == grt::ErrorMsg || msg_type == grt::WarningMsg)
    _exec_sql_error_count++;
  refresh_log_messages(false);
}


void Db_sql_editor::refresh_log_messages(bool ignore_last_message_timestamp)
{
  if (_has_pending_log_messages)
  {
    bool is_refresh_needed= ignore_last_message_timestamp;
    if (!ignore_last_message_timestamp)
    {
      time_t now= timestamp();
      if (_last_log_message_timestamp + _progress_status_update_interval < now)
        is_refresh_needed= true;
      _last_log_message_timestamp= now;
    }
    if (is_refresh_needed)
    {
      exec_sql_task->send_progress(0.f, std::string(), std::string());
      _has_pending_log_messages= false;
    }
  }
}


void Db_sql_editor::get_log_event_details(int log_event_no, int &log_event_type_code, std::string &log_event_time, std::string &log_event_action, std::string &log_event_message)
{
  NodeId node(log_event_no);
  _log->get_field(node, 0, log_event_type_code);
  _log->get_field(node, 2, log_event_time);
  _log->get_field(node, 3, log_event_action);
  _log->get_field(node, 4, log_event_message);
}


int Db_sql_editor::recordset_count()
{
  return _recordsets.size();
}


Recordset::Ref Db_sql_editor::recordset(int index)
{
  return _recordsets.at(index);
}


void Db_sql_editor::init_connection(sql::Connection* dbc_conn_ref, const db_mgmt_ConnectionRef& connectionProperties, sql::Dbc_connection_handler::Ref& dbc_conn)
{
  db_mgmt_RdbmsRef rdbms= db_mgmt_RdbmsRef::cast_from(_connection->driver()->owner());
  SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms);
  Sql_specifics::Ref sql_specifics= sql_facade->sqlSpecifics();

  // connection startup script
  {
    std::list<std::string> sql_script;
    {
      sql_specifics->get_connection_startup_script(sql_script);
      bool use_ansi_quotes= (connectionProperties->parameterValues().get_int("useAnsiQuotes", 0) != 0);
      if (use_ansi_quotes)
      {
        std::string sql= sql_specifics->setting_ansi_quotes();
        if (!sql.empty())
          sql_script.push_back(sql);
      }
    }
    
    // check if SQL_SAFE_UPDATES should be enabled
    if (_options.get_int("DbSqlEditor:SafeUpdates", 1))
      sql_script.push_back("SET SQL_SAFE_UPDATES=1");
    
    std::auto_ptr<sql::Statement> stmt(dbc_conn_ref->createStatement());
    sql::SqlBatchExec sql_batch_exec;
    sql_batch_exec(stmt.get(), sql_script);
  }

  // remember connection id
  {
    std::string query_connection_id= sql_specifics->query_connection_id();
    if (!query_connection_id.empty())
    {
      std::auto_ptr<sql::Statement> stmt(dbc_conn_ref->createStatement());
      stmt->execute(query_connection_id);
      boost::shared_ptr<sql::ResultSet> rs(stmt->getResultSet());
      rs->next();
      dbc_conn->id= rs->getInt(1);
    }
  }
}


void Db_sql_editor::create_connection(sql::Dbc_connection_handler::Ref &dbc_conn, db_mgmt_ConnectionRef db_mgmt_conn, bool autocommit_mode)
{
  dbc_conn->is_disconnection_requested= false;

  sql::DriverManager *dbc_drv_man= sql::DriverManager::getDriverManager();
  dbc_conn->ref= dbc_drv_man->getConnection(db_mgmt_conn,
    sigc::bind(sigc::mem_fun(this, &Db_sql_editor::init_connection), sigc::ref(dbc_conn)));
  
  //! dbms-specific code
  if (dbc_conn->ref->getMetaData()->getDatabaseMajorVersion() < 5)
  {
    throw std::runtime_error("MySQL Server version is older than 5.0, which is not supported");
  }
  
  if (dbc_conn->active_schema.empty())
    dbc_conn->active_schema= dbc_conn->ref->getSchema();
  else
    dbc_conn->ref->setSchema(dbc_conn->active_schema);
  dbc_conn->ref->setAutoCommit(autocommit_mode);
  dbc_conn->autocommit_mode= dbc_conn->ref->getAutoCommit();
}


void Db_sql_editor::connect()
{
  {
    GMutexLock aux_dbc_conn_mutex(_aux_dbc_conn_mutex);
    GMutexLock usr_dbc_conn_mutex(_usr_dbc_conn_mutex);

    reset();

    _aux_dbc_conn->ref.reset();
    _usr_dbc_conn->ref.reset();
    create_connection(_aux_dbc_conn, _connection, _aux_dbc_conn->autocommit_mode);
    create_connection(_usr_dbc_conn, _connection, _usr_dbc_conn->autocommit_mode);

    cache_sql_mode();

    // connection info
    try
    {
      _connection_details["name"] = _connection->name();
      _connection_details["hostName"] = _connection->parameterValues().get_string("hostName");
      _connection_details["port"] = strfmt("%li\n", _connection->parameterValues().get_int("port"));
      _connection_details["socket"] = _connection->parameterValues().get_string("socket");
      _connection_details["driverName"] = _connection->driver()->name();
      _connection_details["dbmsProductName"] = _usr_dbc_conn->ref->getMetaData()->getDatabaseProductName();
      _connection_details["dbmsProductVersion"] = _usr_dbc_conn->ref->getMetaData()->getDatabaseProductVersion();
      _connection_details["userName"] = _connection->parameterValues().get_string("userName");
  
      // Connection:
      _connection_info=       "Connection Information\n";
      _connection_info.append("======================\n");
      _connection_info.append("Name:    ").append(_connection->name()).append("\n");
      // Host:
      if (_connection->driver()->name() == "MysqlNativeSocket")
      {
#ifdef _WIN32
        std::string name = _connection->parameterValues().get_string("socket", "pipe");
#else
        std::string name = _connection->parameterValues().get_string("socket", "UNIX domain socket");
#endif
        _connection_info.append("Localhost through ").append(name).append("\n");
      }
      else
      {
        _connection_info.append("Host:    ").append(
          strfmt("%s:%i\n",
           _connection->parameterValues().get_string("hostName").c_str(),
           (int) _connection->parameterValues().get_int("port")));
      }

      // Server:
      _connection_info.append("Server:  ").append(_connection_details["dbmsProductName"]).append("\n");
      _connection_info.append("Version: ").append(_connection_details["dbmsProductVersion"]).append("\n");
      // User:
      _connection_info.append("User:    ").append(_connection->parameterValues().get_string("userName")).append("\n");
    }
    CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Get connection information"));
  }

// not sure why this was making a copy of rdbms
//  db_mgmt_RdbmsRef rdbms= db_mgmt_RdbmsRef::cast_from(grt::copy_object(_grtm->get_grt(), this->rdbms()));
  db_mgmt_RdbmsRef rdbms= this->rdbms();
  std::string database_package= *rdbms->databaseObjectPackage();

  workbench_physical_ModelRef model(_grtm->get_grt());
  rdbms->owner(model);
  model->rdbms(rdbms);
  db_CatalogRef catalog= _grtm->get_grt()->create_object<db_Catalog>(database_package + ".Catalog");
  {
    catalog->owner(model);
    catalog->name("default");
    catalog->oldName("default");
    catalog->version(rdbms->version());
    grt::replace_contents(catalog->simpleDatatypes(), rdbms->simpleDatatypes());

    // setup list of synonyms
    grt::Module *module= _grtm->get_grt()->get_module("DbMySQL");
    if (module)
    {
      grt::BaseListRef args(_grtm->get_grt());
      args.ginsert(rdbms);
      grt::ListRef<db_UserDatatype> user_types(grt::ListRef<db_UserDatatype>::cast_from(module->call_function("getDefaultUserDatatypes", args)));
      
      if (user_types.is_valid())
      {
        GRTLIST_FOREACH(db_UserDatatype, user_types, ut)
        {
          (*ut)->owner(catalog);
          catalog->userDatatypes().insert(*ut);
        }
      }
    }    
    
    model->catalog(catalog);
  }
 
  {
    GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();
    _live_physical_overview->set_model(model);
  }
  
  _schema_tree.set_active_schema(_connection->parameterValues().get_string("schema"));
  
  live_schemata_refresh_task->execute_in_main_thread(
    sigc::bind(sigc::mem_fun(this, (void(Db_sql_editor::*)(bool))&Db_sql_editor::refresh_physical_overview), false),
    false,
    true);
    
  update_menu_and_toolbar();
}


bool Db_sql_editor::connected() const
{
  return (_usr_dbc_conn && _usr_dbc_conn->ref.get_ptr() && !_usr_dbc_conn->ref->isClosed());
}


GMutexLock Db_sql_editor::ensure_valid_aux_connection()
{
  return ensure_valid_dbc_connection(_aux_dbc_conn, _aux_dbc_conn_mutex);
}


GMutexLock Db_sql_editor::ensure_valid_usr_connection()
{
  return ensure_valid_dbc_connection(_usr_dbc_conn, _usr_dbc_conn_mutex);
}


void Db_sql_editor::close_connection(sql::Dbc_connection_handler::Ref &dbc_conn)
{
  if (dbc_conn && dbc_conn->ref.get_ptr())
  {
    try
    {
      dbc_conn->ref->close();
    }
    catch (sql::SQLException &)
    {
      // ignore if the connection is already closed
    }
  }
}


GMutexLock Db_sql_editor::ensure_valid_dbc_connection(sql::Dbc_connection_handler::Ref &dbc_conn, GMutex *dbc_conn_mutex)
{
  GMutexLock mutex_lock(dbc_conn_mutex);
  bool valid= false;
  if (dbc_conn && dbc_conn->ref.get_ptr())
  {
    if (dbc_conn->ref->isClosed())
    {
      if (dbc_conn->autocommit_mode)
      {
        create_connection(dbc_conn, _connection, dbc_conn->autocommit_mode);
        if (!dbc_conn->ref->isClosed())
          valid= true;
      }
    }
    else
      valid= true;
  }
  if (!valid)
    throw std::runtime_error("DBMS connection is not available");

  return mutex_lock;
}


bool Db_sql_editor::auto_commit()
{
  if (_usr_dbc_conn)
    return _usr_dbc_conn->autocommit_mode;
  return false;
}


void Db_sql_editor::auto_commit(bool value)
{
  if (!_usr_dbc_conn)
    return;
  {
    const char *STATEMENT= value ? "AUTOCOMMIT=1" : "AUTOCOMMIT=0";
    try
    {
      GMutexLock usr_dbc_conn_mutex= ensure_valid_usr_connection();
      _usr_dbc_conn->ref->setAutoCommit(value);
      _usr_dbc_conn->autocommit_mode= _usr_dbc_conn->ref->getAutoCommit();
    }
    CATCH_ANY_EXCEPTION_AND_DISPATCH(STATEMENT)
  }
  update_menu_and_toolbar();
}


void Db_sql_editor::cancel_query()
{
  std::string query_kill_connection;
  {
    db_mgmt_RdbmsRef rdbms= db_mgmt_RdbmsRef::cast_from(_connection->driver()->owner());
    SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms);
    Sql_specifics::Ref sql_specifics= sql_facade->sqlSpecifics();
    query_kill_connection= sql_specifics->query_kill_connection(_usr_dbc_conn->id);
  }
  if (query_kill_connection.empty())
    return;

  const char *STATEMENT= "INTERRUPT";
  RowId log_message_index= add_log_message(grt::InfoMsg, _("Running..."), STATEMENT, "");
  Timer timer(false);

  try
  {
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      std::auto_ptr<sql::Statement> stmt(_aux_dbc_conn->ref->createStatement());
      {
        ScopeExitTrigger schedule_timer_stop(sigc::mem_fun(&timer, &Timer::stop));
        timer.run();
        stmt->execute(query_kill_connection);
        
        // this can potentially cause threading issues, since connector driver isn't thread-safe
        //close_connection(_usr_dbc_conn);

        // connection drop doesn't interrupt fetching stage (surprisingly)
        // to workaround that we set special flag and check it periodically during fetching
        _usr_dbc_conn->is_disconnection_requested= true;
      }
    }

    set_log_message(log_message_index, grt::InfoMsg, _("OK"), STATEMENT, timer.duration_formatted());

    // reconnect but only if in autocommit mode
    if (_usr_dbc_conn->autocommit_mode)
    {
      // this will restore connection if it was established previously
      exec_sql_task->execute_in_main_thread(
        sigc::mem_fun(this, &Db_sql_editor::send_message_keep_alive),
        false,
        true);
    }
  }
  CATCH_SQL_EXCEPTION_AND_DISPATCH(STATEMENT, log_message_index, "")
}


void Db_sql_editor::commit()
{
  exec_sql_retaining_editor_contents("COMMIT", false);
}


void Db_sql_editor::rollback()
{
  exec_sql_retaining_editor_contents("ROLLBACK", false);
}


bool Db_sql_editor::explain_sql(bool only_check_if_applicable)
{
  int start, end;
  Sql_editor::Ref sql_editor_= sql_editor();
  sql_editor_->selected_range(start, end);
  std::string sql= sql_editor_->sql();
  if (start != end)
    sql= sql.substr(start, end-start);

  do_explain_sql(sql);
  return true;
}


void Db_sql_editor::explain_current_statement()
{
  Sql_editor::Ref sql_editor_= sql_editor();
  do_explain_sql(sql_editor_->current_statement());
}


void Db_sql_editor::do_explain_sql(const std::string &sql)
{
  SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms());
  std::list<std::string> statements;
  sql_facade->splitSqlScript(sql, statements);
  Sql_syntax_check::Ref sql_syntax_check= sql_facade->sqlSyntaxCheck();
  std::string sql_script;
  for (std::list<std::string>::iterator i= statements.begin(), i_end= statements.end(); i != i_end; ++i)
    if (Sql_syntax_check::sql_select == sql_syntax_check->determine_statement_type(*i))
      sql_script+= "EXPLAIN " + *i + ";\n";

  exec_sql_retaining_editor_contents(sql_script, false);
}


void Db_sql_editor::exec_sql_retaining_editor_contents(const std::string &sql_script, bool sync, bool dont_add_limit_clause)
{
  auto_save();
  
  exec_sql_task->exec(sync,
    sigc::bind(sigc::mem_fun(this, &Db_sql_editor::do_exec_sql),
               weak_ptr_from(this), sql_script, Sql_editor::Ref(),
               (ExecFlags)(Retaining | (dont_add_limit_clause?Dont_add_limit_clause:0)),
               &_recordsets));
}


void Db_sql_editor::set_sql_editor_text(const std::string &sql, bool run_sql)
{
  sql_editor()->is_refresh_enabled(true);
  sql_editor()->sql(sql);
  do_partial_ui_refresh(Db_sql_editor::RefreshEditor);
  if (run_sql)
    do_partial_ui_refresh(Db_sql_editor::RunCurrentScript);
}


void Db_sql_editor::on_sql_editor_text_change()
{
  if (_updating_sql_editor == 0)
  {
    sql_editor_dirty(_active_sql_editor_index, true);
  }
}


void Db_sql_editor::on_sql_editor_text_selection_change()
{
  // not necessary in mac, maybe in windows or linux.. if useless, just remove it
//  _context_ui->signal_validate _edit_menu_items().emit();
}


Db_sql_editor::Recordsets Db_sql_editor::exec_sql_returning_results(const std::string &sql_script, bool dont_add_limit_clause)
{
  Recordsets rsets;
  std::string sql(sql_script);
  
  do_exec_sql(_wbsql->get_grt_manager()->get_grt(), weak_ptr_from(this), sql, Sql_editor::Ref(), 
              (ExecFlags)(dont_add_limit_clause?Dont_add_limit_clause:0),
              &rsets);
  
  return rsets;
}

void Db_sql_editor::exec_sql(std::string &sql, Sql_editor::Ref editor, bool sync, bool wrap_with_non_std_delimiter, bool dont_add_limit_clause)
{
  auto_save();
  
  exec_sql_task->exec(sync,
    sigc::bind(sigc::mem_fun(this, &Db_sql_editor::do_exec_sql),
               weak_ptr_from(this), sql, editor, 
               (ExecFlags)((wrap_with_non_std_delimiter ? Wrap_with_non_std_delimiter : 0)|(dont_add_limit_clause?Dont_add_limit_clause:0)),
               &_recordsets));
}


grt::StringRef Db_sql_editor::do_exec_sql(grt::GRT *grt, Ptr self_ptr, std::string &sql, Sql_editor::Ref editor, ExecFlags flags, Recordsets *result_list)
{
  bool retaining = (flags & Retaining) != 0;
  bool wrap_with_non_std_delimiter = (flags & Wrap_with_non_std_delimiter) != 0;
  bool dont_add_limit_clause = (flags & Dont_add_limit_clause) != 0;
  int active_recordset_index = -1;
  
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Db_sql_editor, self_ptr, self, grt::StringRef(""))

  // add_log_message() will increment this variable on errors or warnings
  _exec_sql_error_count = 0;

  sql::Driver *dbc_driver= NULL;
  std::string statement;
  try
  {/*
    if (!retaining)
    {
      if (!_recordsets_are_pinned_by_default)
      {
        GMutexLock recordsets_mutex(_recordsets_mutex);
        Recordsets recordsets;
        recordsets.reserve(result_list->size());
        BOOST_FOREACH (Recordset::Ref rs, *result_list)
        {
          if (rs->pinned() || !rs->can_close(false))
            recordsets.push_back(rs);
          else
            recordset_list_changed.emit(rs, false);
        }
        result_list->swap(recordsets);        
      }
    }*/
    Recordsets::iterator recordset_insertion_point;
    if (editor && !retaining && !_recordsets_are_pinned_by_default)
    {
      Sql_editors::iterator info;
      for (info = _sql_editors.begin(); info != _sql_editors.end(); ++info)
      {
        if (info->editor == editor)
          break;
      }
      
      if (info != _sql_editors.end())
      {
        std::vector<long long>::iterator iter;
        iter = std::find(info->resultsets.begin(), info->resultsets.end(), info->active_resultset);
        if (iter != info->resultsets.end())
          active_recordset_index = iter - info->resultsets.begin();
      }
      
      // close recordsets that are attached to the editor and leave the rest
      GMutexLock recordsets_mutex(_recordsets_mutex);
      Recordsets recordsets;
      recordsets.reserve(result_list->size());
      BOOST_FOREACH (Recordset::Ref rs, *result_list)
      {
        bool this_editor = false;
        std::vector<long long>::iterator iter;
        if (info != _sql_editors.end() && (iter = std::find(info->resultsets.begin(), info->resultsets.end(), rs->key())) != info->resultsets.end())
          this_editor = true;

        if (!this_editor || rs->pinned() || !rs->can_close(false))
          recordsets.push_back(rs);
        else
        {
          if (iter != info->resultsets.end())
            info->resultsets.erase(iter);
          recordset_list_changed.emit(rs, false);
          //delme close_recordset_ui.emit(rs->key());
        }
      }
      result_list->swap(recordsets);
      
      recordset_insertion_point = result_list->end();
      // find the 1st recordset of the editor next to this one
      if (info != _sql_editors.end())
      {
        while (++info != _sql_editors.end() && info->resultsets.empty());

        if (info != _sql_editors.end())
        {
          long long first_of_next = info->resultsets.front();
          for (Recordsets::iterator iter = result_list->begin(); iter != result_list->end(); ++iter)
          {
            if (first_of_next == (*iter)->key())
            {
              recordset_insertion_point = iter;
              break;
            }
          }
        }
      }
    }
    else
      recordset_insertion_point = result_list->end();

    GMutexLock use_dbc_conn_mutex= ensure_valid_usr_connection();

    dbc_driver= _usr_dbc_conn->ref->getDriver();
    dbc_driver->threadInit();

    bool is_running_query= true;
    AutoSwap<bool> is_running_query_keeper(_is_running_query, is_running_query);
    update_menu_and_toolbar();

    _has_pending_log_messages= false;
    bec::TimerActionThread *log_messages_refresh_thread= TimerActionThread::create(
      sigc::bind(sigc::mem_fun(this, &Db_sql_editor::refresh_log_messages), true),
      _progress_status_update_interval);
    ScopeExitTrigger schedule_log_messages_refresh_thread_stop(sigc::bind(
      sigc::mem_fun(log_messages_refresh_thread, &bec::TimerActionThread::stop), false));
    ScopeExitTrigger schedule_log_messages_refresh(sigc::bind(
      sigc::mem_fun(this, &Db_sql_editor::refresh_log_messages), true));

    SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms());
    Sql_syntax_check::Ref sql_syntax_check= sql_facade->sqlSyntaxCheck();
    Sql_specifics::Ref sql_specifics= sql_facade->sqlSpecifics();

    if (wrap_with_non_std_delimiter)
      sql= sql_specifics->setting_non_std_sql_delimiter() + sql;

    //XXX this will create a copy of the whole script, should work with ranges instead
    std::list<std::string> statements;
    sql_facade->splitSqlScript(sql, statements);
    _history->add_entry(statements);
    std::list<Recordset::Ref> added_recordsets;
    BOOST_FOREACH (statement, statements)
    {
      std::list<std::string> sub_statements;
      sql_facade->splitSqlScript(statement, sub_statements);
      int multiple_statement_count= sub_statements.size();
      bool is_multiple_statement= (1 < multiple_statement_count);
      {
        statement= strip_text(statement, false, true);
        if (statement.empty())
          continue;

        Sql_syntax_check::Statement_type statement_type= sql_syntax_check->determine_statement_type(statement);
        if (Sql_syntax_check::sql_empty == statement_type)
          continue;

        std::string schema_name;
        std::string table_name;
        std::string editable_query_tail;
        if (!is_multiple_statement && (Sql_syntax_check::sql_edit == statement_type))
        {
          // to open table contents in edit mode the following syntax is used: EDIT [SCHEMA.]TABLE ...
          // this type of query needs special preprocessing: rewriting using 'select' statement
          statement_type= Sql_syntax_check::sql_select;
          sql_syntax_check->parse_edit_statement(statement, schema_name, table_name, editable_query_tail);
          if (schema_name.empty())
            schema_name= active_schema();
          if (schema_name.empty())
            schema_name= _usr_dbc_conn->ref->getSchema();
        }

        Recordset_cdbc_storage::Ref data_storage;

        // for select queries add limit clause if specified by global option
        if (!is_multiple_statement && (Sql_syntax_check::sql_select == statement_type))
        {
          data_storage= Recordset_cdbc_storage::create(_grtm);
          data_storage->rdbms(rdbms());
          data_storage->dbms_conn(_usr_dbc_conn);
          if (table_name.empty())
            data_storage->sql_query(statement);
          data_storage->schema_name(schema_name);
          data_storage->table_name(table_name);
          data_storage->additional_clauses(editable_query_tail);
          {
            bool limit_rows= !dont_add_limit_clause && (0 != _options.get_int("SqlEditor:LimitRows", 0));
            data_storage->limit_rows(limit_rows);

            int limit_rows_count= _options.get_int("SqlEditor:LimitRowsCount");
            if (0 != limit_rows_count)
              data_storage->limit_rows_count(limit_rows_count);
          }
          statement= data_storage->decorated_sql_query();
        }

        {
          RowId log_message_index= add_log_message(grt::InfoMsg, _("Running..."), statement,
            ((Sql_syntax_check::sql_select == statement_type) ? "? / ?" : "?"));

          bool statement_failed= false;
          long long updated_rows_count= -1;
          Timer statement_exec_timer(false);
          Timer statement_fetch_timer(false);
          boost::shared_ptr<sql::Statement> dbc_statement(_usr_dbc_conn->ref->createStatement());
          bool is_result_set_first= false;
          try
          {
            {
              ScopeExitTrigger schedule_statement_exec_timer_stop(sigc::mem_fun(&statement_exec_timer, &Timer::stop));
              statement_exec_timer.run();
              is_result_set_first= dbc_statement->execute(statement);
            }
            updated_rows_count= dbc_statement->getUpdateCount();
            if (Sql_syntax_check::sql_use == statement_type)
              cache_active_schema_name();
          }
          catch (sql::SQLException &e)
          {
            std::string err_msg;
            // safe mode
            if (e.getErrorCode() == 1175)
              err_msg= strfmt(_("Error Code: %i. To disable safe mode, toggle the option in Preferences -> SQL Editor -> Query Editor and reconnect.\n%s"), e.getErrorCode(), e.what());
            else
              err_msg= strfmt(_("Error Code: %i\n%s"), e.getErrorCode(), e.what());
            set_log_message(log_message_index, grt::ErrorMsg, err_msg, statement, statement_exec_timer.duration_formatted());
            statement_failed= true;
          }
          catch (std::exception &e)
          {
            std::string err_msg= strfmt(_("Error: %s"), e.what());
            set_log_message(log_message_index, grt::ErrorMsg, err_msg, statement, statement_exec_timer.duration_formatted());
            statement_failed= true;
          }
          if (statement_failed)
          {
            if (_continue_on_error)
              continue; // goto next statement
            else
              goto stop_processing_sql_script;
          }

          sql::mysql::MySQL_Connection* mysql_connection = dynamic_cast<sql::mysql::MySQL_Connection*>(dbc_statement->getConnection());
          sql::SQLString last_statement_info;
          if (mysql_connection != NULL)
            last_statement_info = mysql_connection->getLastStatementInfo();
          if (updated_rows_count >= 0)
          {
            std::string message = strfmt(_("%lli row(s) affected"), updated_rows_count);
            if (!last_statement_info->empty())
              message.append("\n").append(last_statement_info);
            set_log_message(log_message_index, grt::InfoMsg, message, statement, statement_exec_timer.duration_formatted());
          }

          int resultset_count= 0;
          bool more_results= is_result_set_first;
          bool reuse_log_msg= false;
          if ((updated_rows_count < 0) || is_multiple_statement)
          {
            for (int processed_substatements_count= 0; processed_substatements_count < multiple_statement_count; ++processed_substatements_count)
            {
              do
              {
                if (more_results)
                {
                  if (!reuse_log_msg && ((updated_rows_count >= 0) || (resultset_count)))
                    log_message_index= add_log_message(grt::InfoMsg, _("Fetching..."), statement, "- / ?");
                  else
                    set_log_message(log_message_index, grt::InfoMsg, _("Fetching..."), statement, statement_exec_timer.duration_formatted() + " / ?");
                  reuse_log_msg= false;
                  boost::shared_ptr<sql::ResultSet> dbc_resultset;
                  {
                    ScopeExitTrigger schedule_statement_fetch_timer_stop(sigc::mem_fun(&statement_fetch_timer, &Timer::stop));
                    statement_fetch_timer.run();
                    dbc_resultset.reset(dbc_statement->getResultSet());
                  }
                  if (dbc_resultset)
                  {
                    if (!data_storage)
                    {
                      data_storage= Recordset_cdbc_storage::create(_grtm);
                      data_storage->rdbms(rdbms());
                      data_storage->dbms_conn(_usr_dbc_conn);
                      if (table_name.empty())
                        data_storage->sql_query(statement);
                      data_storage->schema_name(schema_name);
                      data_storage->table_name(table_name);
                    }

                    data_storage->dbc_statement(dbc_statement);
                    data_storage->dbc_resultset(dbc_resultset);
                    data_storage->reloadable(!is_multiple_statement && (Sql_syntax_check::sql_select == statement_type));

                    Recordset::Ref rs= Recordset::create(exec_sql_task);
                    rs->is_field_value_truncation_enabled(true);
                    rs->apply_changes= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::apply_changes_to_recordset), Recordset::Ptr(rs));
                    rs->on_close.connect(sigc::mem_fun(*this, &Db_sql_editor::on_close_recordset));
                    rs->caption(strfmt("%s (%i)",
                                       (table_name.empty() ? _("Result") : table_name.c_str()),
                                       ++_rs_sequence));
                    rs->generator_query(statement);
                    rs->action_list().register_action("recall_query",
                      sigc::bind(sigc::mem_fun(this, &Db_sql_editor::recall_recordset_query), Recordset::Ptr(rs)));
                    rs->refresh_ui_status_bar_signal.connect(sigc::hide_return(
                      sigc::bind(sigc::mem_fun(this, &bec::RefreshUI::do_partial_ui_refresh), (int)RefreshRecordsetTitle)));
                    rs->data_storage(data_storage);
                    rs->export_wizard= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::show_export_recordset), Recordset::Ptr(rs));
                    rs->popup_menu_items_slot= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::recordset_popup_menu_items), rs);
                    rs->call_popup_menu_item_slot= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::call_recordset_popup_menu_item), rs);
                    rs->reset(true);
                    rs->status_text_trailer = strfmt("Duration: %s, fetched in: %s",
                                                     statement_exec_timer.duration_formatted().c_str(),
                                                     statement_fetch_timer.duration_formatted().c_str());                                                     

                    if (data_storage->valid()) // query statement
                    {
                      {
                        GMutexLock recordsets_mutex(_recordsets_mutex);
                        recordset_insertion_point = result_list->insert(recordset_insertion_point, rs);
                        recordset_insertion_point++;
                      }
                      recordset_list_changed.emit(rs, true);
                      std::string statement_res_msg= strfmt(_("%i row(s) returned"), rs->row_count());
                      if (!last_statement_info->empty())
                        statement_res_msg.append("\n").append(last_statement_info);
                      std::string exec_and_fetch_durations=
                        (((updated_rows_count >= 0) || (resultset_count)) ? std::string("-") : statement_exec_timer.duration_formatted()) + " / " +
                        statement_fetch_timer.duration_formatted();
                      set_log_message(log_message_index, grt::InfoMsg, statement_res_msg, statement, exec_and_fetch_durations);
                    }
                    //! else failed to fetch data
                    added_recordsets.push_back(rs);

                    ++resultset_count;
                  }
                  else
                  {
                    reuse_log_msg= true;
                  }
                  data_storage.reset();
                }
              }
              while ((more_results= dbc_statement->getMoreResults()));
            }
          }
          
          if ((updated_rows_count < 0) && !(resultset_count))
          {
            set_log_message(log_message_index, grt::InfoMsg, _("OK"), statement, statement_exec_timer.duration_formatted());
          }
        }
      }
    } // BOOST_FOREACH (statement, statements)

    // recaption resultsets to match the editor and add them to the list of resultsets for the ditor
    if (editor)
    {
      GMutexLock lock(_sql_editors_mutex);
      
      for (Sql_editors::iterator iter = _sql_editors.begin(); iter != _sql_editors.end(); ++iter)
      {
        if (iter->editor == editor)
        {
          if (iter->resultsets.empty() && added_recordsets.size() == 1)
          {
            added_recordsets.front()->caption(strfmt("%s Result", iter->caption.c_str()));
            iter->resultsets.push_back(added_recordsets.front()->key());
          }
          else
          {
            for (std::list<Recordset::Ref>::iterator rs= added_recordsets.begin();
                 rs != added_recordsets.end(); ++rs)
            {
              iter->resultsets.push_back((*rs)->key());
              (*rs)->caption(strfmt("%s Result %i", iter->caption.c_str(), (int)iter->resultsets.size()));
            }
          }
          
          if (active_recordset_index >= 0 && active_recordset_index < (int)iter->resultsets.size())
            iter->active_resultset = iter->resultsets[active_recordset_index];
          else
            if (!iter->resultsets.empty())
              iter->active_resultset = iter->resultsets.back();
          break;
        }
      }
    }

stop_processing_sql_script:
    cache_sql_mode();
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH(statement)

  if (dbc_driver)
    dbc_driver->threadEnd();

  update_menu_and_toolbar();
  
  return grt::StringRef("");
}


bool Db_sql_editor::is_running_query()
{
  return _is_running_query;
}


void Db_sql_editor::recordsets_are_pinned_by_default(bool value)
{
  _recordsets_are_pinned_by_default= value; 
  
  update_menu_and_toolbar();
}


void Db_sql_editor::continue_on_error(bool val)
{
  if (_continue_on_error == val)
    return;

  _continue_on_error= val;
  _options.set("DbSqlEditor:ContinueOnError", grt::IntegerRef((int)_continue_on_error));
  
  if (_menu)
    _menu->set_item_checked("query.stopOnError", !continue_on_error());
  if (_toolbar)
    _toolbar->set_item_checked("query.stopOnError", !continue_on_error());
}


void Db_sql_editor::send_message_keep_alive()
{
  try
  {
    // ping server and reset connection timeout counter
    // this also checks the connection state and restores it if possible
    ensure_valid_aux_connection();
    ensure_valid_usr_connection();
  }
  catch (const std::exception &)
  {
  }
}


void Db_sql_editor::active_recordset(Recordset::Ref rset)
{
  _active_recordset= rset;
  
  if (rset)
  {
    GMutexLock lock(_sql_editors_mutex);
    int editor = sql_editor_index_for_recordset(rset->key());
    if (editor >= 0)
    {
      active_sql_editor_index(editor);
      _sql_editors[editor].active_resultset = rset->key();
    }
  }

  if (_menu)
  {
    _menu->set_item_enabled("query.save_edits", rset && rset->has_pending_changes());
    _menu->set_item_enabled("query.discard_edits", rset && rset->has_pending_changes());
    _menu->set_item_enabled("query.export", (bool)rset);
  }
}


void Db_sql_editor::on_close_recordset(Recordset::Ptr rs_ptr)
{
  bool got_empty = false;
  int editor;
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs)
  {
    {
      GMutexLock ed_lock(_sql_editors_mutex);
      editor = sql_editor_index_for_recordset(rs_ref->key());
      if (editor >= 0)
      {
        std::vector<long long> &rsets(_sql_editors[editor].resultsets);
        rsets.erase(std::find(rsets.begin(), rsets.end(), rs_ref->key()));
        if (rsets.empty())
          got_empty = true;
      }
    }
    
    GMutexLock recordsets_mutex(_recordsets_mutex);
    Recordsets::iterator i= std::find(_recordsets.begin(), _recordsets.end(), rs_ref);
    if (_recordsets.end() != i)
      _recordsets.erase(i);
    recordset_list_changed(rs_ref, false);
    //delme close_recordset_ui.emit(rs->key());    
  }
  if (got_empty)
    add_empty_recordset_for_sql_editor(editor);
}

void Db_sql_editor::recall_recordset_query(Recordset::Ptr rs_ptr)
{
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs)
  {
    std::string query = rs->generator_query();
    
    new_sql_scratch_area();
    set_sql_editor_text(query, false);
  }
}

void Db_sql_editor::save_active_recordset_edits()
{
  Recordset::Ref rs= _active_recordset;
  if (rs)
    rs->apply_changes();
  
  if (_menu)
  {
    _menu->set_item_enabled("query.save_edits", rs && rs->has_pending_changes());
    _menu->set_item_enabled("query.discard_edits", rs && rs->has_pending_changes());
  }
}


void Db_sql_editor::discard_active_recordset_edits()
{
  Recordset::Ref rs= _active_recordset;
  if (rs)
    rs->rollback();
  
  if (_menu)
  {
    _menu->set_item_enabled("query.save_edits", rs && rs->has_pending_changes());
    _menu->set_item_enabled("query.discard_edits", rs && rs->has_pending_changes());
  }
}


void Db_sql_editor::apply_changes_to_recordset(Recordset::Ptr rs_ptr)
{
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs)

  try
  {
    GMutexLock usr_dbc_conn_mutex= ensure_valid_usr_connection();

    // we need transaction to enforce atomicity of change set
    // so if autocommit is currently enabled disable it temporarily
    bool auto_commit= _usr_dbc_conn->ref->getAutoCommit();
    ScopeExitTrigger autocommit_mode_keeper;
    int res= -2;
    if (!auto_commit)
    {
      int res= mforms::Utilities::show_warning(
        _("Apply Changes to Recordset"),
        _("Autocommit is currently disabled. Do you want to perform a COMMIT before applying the changes?\n"
          "If you do not commit, a failure during the recordset update will result in a rollback of the active transaction, if you have one."),
        _("Commit and Apply"),
        _("Cancel"),
        _("Apply"));

      if (res == mforms::ResultOk)
      {
        _usr_dbc_conn->ref->commit();
      }
    }
    else
    {
      autocommit_mode_keeper.slot= sigc::bind(
        sigc::mem_fun(_usr_dbc_conn->ref.get(), &sql::Connection::setAutoCommit),
        auto_commit);
      _usr_dbc_conn->ref->setAutoCommit(false);
    }

    if (res != mforms::ResultCancel) // only if not canceled
    {
      bool is_data_changes_commit_wizard_enabled= (0 != _options.get_int("DbSqlEditor:IsDataChangesCommitWizardEnabled", 1));
      if (is_data_changes_commit_wizard_enabled)
      {
        /*bool are_changes_committed= */run_data_changes_commit_wizard(rs_ptr);
      }
      else
      {
        rs->apply_changes_();
      }
    }
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Apply changes to recordset"))
}


bool Db_sql_editor::run_data_changes_commit_wizard(Recordset::Ptr rs_ptr)
{
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs, false)

  on_sql_script_run_error.clear();
  on_sql_script_run_progress.clear();
  on_sql_script_run_statistics.clear();

  // set underlying recordset data storage to use sql substitute (potentially modified by user)
  // instead of generating sql based on swap db contents
  Recordset_data_storage::Ref data_storage_ref= rs->data_storage();
  Recordset_sql_storage *sql_storage= dynamic_cast<Recordset_sql_storage *>(data_storage_ref.get());
  if (!sql_storage)
    return false;
  sql_storage->init_sql_script_substitute(rs_ptr, true);
  sql_storage->is_sql_script_substitute_enabled(true);
  const Sql_script &sql_script= sql_storage->sql_script_substitute();;
  std::string sql_script_text= Recordset_sql_storage::statements_as_sql_script(sql_script.statements);

  SqlScriptRunWizard wizard(_grtm);
  on_sql_script_run_error.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_error));
  on_sql_script_run_progress.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_exec_progress));
  on_sql_script_run_statistics.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_exec_stat));
  wizard.values().gset("sql_script", sql_script_text);
  wizard.apply_page->apply_sql_script= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::apply_data_changes_commit), rs_ptr);
  wizard.run_modal();

  return !wizard.has_errors();
}


void Db_sql_editor::apply_data_changes_commit(std::string &sql_script_text, Recordset::Ptr rs_ptr)
{
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs);

  // this lock is supposed to be acquired lower in call-stack by Db_sql_editor::apply_changes_to_recordset
  //GMutexLock usr_conn_mutex= ensure_valid_usr_connection();

  Recordset_data_storage::Ref data_storage_ref= rs->data_storage();
  Recordset_sql_storage *sql_storage= dynamic_cast<Recordset_sql_storage *>(data_storage_ref.get());
  if (!sql_storage)
    return;

  Sql_script sql_script= sql_storage->sql_script_substitute();
  sql_script.statements.clear();
  SqlFacade::Ref sql_splitter= SqlFacade::instance_for_rdbms(rdbms());
  sql_splitter->splitSqlScript(sql_script_text, sql_script.statements);

  sigc::connection on_sql_script_run_error_conn= sql_storage->on_sql_script_run_error.connect(on_sql_script_run_error.make_slot());
  sigc::connection on_sql_script_run_progress_conn= sql_storage->on_sql_script_run_progress.connect(on_sql_script_run_progress.make_slot());
  sigc::connection on_sql_script_run_statistics_conn= sql_storage->on_sql_script_run_statistics.connect(on_sql_script_run_statistics.make_slot());

  try
  {
    sql_storage->sql_script_substitute(sql_script);
    rs->do_apply_changes(_grtm->get_grt(), rs_ptr, Recordset_data_storage::Ptr(data_storage_ref));
  }
  catch (...)
  {
    on_sql_script_run_error_conn.disconnect();
    on_sql_script_run_progress_conn.disconnect();
    on_sql_script_run_statistics_conn.disconnect();
    throw;
  }
  on_sql_script_run_error_conn.disconnect();
  on_sql_script_run_progress_conn.disconnect();
  on_sql_script_run_statistics_conn.disconnect();

  _history->add_entry(sql_script.statements);
}


void Db_sql_editor::show_export_recordset(Recordset::Ptr rs_ptr)
{
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs)
  RecordsetExportForm exporter(this, rs_ref);
  exporter.run();  
}

void Db_sql_editor::new_sql_script_file()
{
  _active_sql_editor_index= add_sql_editor(false);
  sql_editor_new_ui.emit(_active_sql_editor_index);
}


void Db_sql_editor::new_sql_scratch_area()
{
  _active_sql_editor_index= add_sql_editor(true);
  sql_editor_new_ui.emit(_active_sql_editor_index);
}


void Db_sql_editor::revert_sql_script_file()
{
  const Sql_editor_info &info(_sql_editors[_active_sql_editor_index]);
  
  info.editor->sql("");
  open_sql_script_file(info.filename, false, false, info.orig_encoding);
  
  _grtm->replace_status_text(strfmt(_("Reverted to saved '%s'."), info.filename.c_str()));
}


void Db_sql_editor::open_sql_script_file(const std::string &file_path, bool run_immediately, bool run_only_skip_editor,
                                         const std::string &encoding)
{
  if (file_path.empty())
    return;
  
  if (!run_only_skip_editor)
  {
    if (_active_sql_editor_index < 0 || _active_sql_editor_index >= (int) _sql_editors.size() ||
        !_sql_editors[_active_sql_editor_index].editor->sql().empty())
      new_sql_script_file();
  }

  gchar *contents= NULL;
  gsize length= 0;
  GError *error= NULL;
  
  _grtm->replace_status_text(strfmt(_("Loading SQL script file '%s'..."), file_path.c_str()));
  
  if (!g_file_get_contents(file_path.c_str(), &contents, &length, &error))
  {
    _grtm->replace_status_text(strfmt(_("Error loading SQL script file '%s'."), file_path.c_str()));
    mforms::Utilities::show_error(strfmt(_("Error opening file %s"), file_path.c_str()), error->message, _("OK"));
    g_error_free(error);
    return;
  }
    
  std::string utf8_contents;
  std::string utf8_contents_fixed;
  std::string original_encoding;
  if (!FileCharsetDialog::ensure_filedata_utf8(contents, length, encoding,
                                               file_path, utf8_contents, &original_encoding))
  {
    g_free(contents);
    _grtm->replace_status_text(_("Cancelled"));
    return;
  }
  g_free(contents);

  if (run_only_skip_editor)
  {    
    _grtm->replace_status_text(strfmt(_("Executing SQL script file '%s'..."), file_path.c_str()));
    
    exec_sql_retaining_editor_contents(utf8_contents, true); 
    
    _grtm->replace_status_text(strfmt(_("Finished executing SQL script file '%s'."), file_path.c_str()));
    
    if (_exec_sql_error_count == 0)
      mforms::Utilities::show_message(_("Execute Script File"),
                                      _("Script file execution finished with no errors."), "OK", "", "");
    else
      mforms::Utilities::show_message(_("Execute Script File"),
                                      strfmt(_("Script file execution finished with %i errors."), _exec_sql_error_count),
                                      "OK", "", "");
    return;
  }  
  
  std::string *sql= NULL;
  if (EolHelpers::check(utf8_contents))
  {
    sql= &utf8_contents;
  }
  else
  {
    EolFormatDialog dialog(_("Inconsistent Line Endings"),
      strfmt(_("File: \"%s\"\nLine endings are inconsistent.\nSelect line ending format to apply.\nResult of conversion will replace original file contents if file is saved afterwards."),
        file_path.c_str()));
    if (dialog.run())
    {
      EolHelpers::fix(utf8_contents, utf8_contents_fixed, dialog.eol_format());
      sql= &utf8_contents_fixed;
    }
  }

  if (!sql) // line endings fix was cancelled by user
    return;

  _updating_sql_editor++;
  set_sql_editor_text(*sql, run_immediately);
  _updating_sql_editor--;
  
  _sql_editors[_active_sql_editor_index].filename = file_path;
  _sql_editors[_active_sql_editor_index].orig_encoding = original_encoding;
  _sql_editors[_active_sql_editor_index].dirty = false;
  _sql_editors[_active_sql_editor_index].is_scratch = false;
  _context_ui->get_wb()->add_recent_file(file_path);

  do_partial_ui_refresh(RefreshEditorTitle);

  _grtm->replace_status_text(strfmt(_("Loaded SQL script file '%s'."), file_path.c_str()));
}


// file_name - filename to save to or empty if it should be asked to user
// editor_index - index of the editor which contents should be saved
// Returns true if file was saved or false if it was cancelled or failed.
bool Db_sql_editor::save_sql_script_file(const std::string &file_name, int editor_index)
{  
  std::string path = file_name;
  
  if (path.empty())
  {
    mforms::FileChooser dlg(mforms::SaveFile);
    
    dlg.set_title(_("Save SQL Script"));
    dlg.set_extensions("SQL Files (*.sql)|*.sql", "sql");
    if (!dlg.run_modal())
      return false;
    
    path = dlg.get_path();
  }
  
  if (!path.empty())
  {
    GError *error= NULL;
  
// this is already done in FileChooser  
//    if (!g_str_has_suffix(path.c_str(), ".sql"))
//      path.append(".sql");
    
    // tell UI that editor data changed
    do_partial_ui_refresh(Db_sql_editor::RefreshEditorBackend);
    
    std::string contents= sql_editor(editor_index)->sql();
    
    _grtm->replace_status_text(strfmt(_("Saving SQL script to '%s'..."), path.c_str()));
    
    if (!g_file_set_contents(path.c_str(), contents.data(), contents.length(), &error))
    {
      _grtm->replace_status_text(strfmt(_("Error saving SQL script to '%s'."), path.c_str()));
      
      mforms::Utilities::show_error(strfmt(_("Error writing file %s"), path.c_str()),
                                    error->message, _("OK"));
      g_error_free(error);
      return false;
    }
    
    _sql_editors[editor_index].filename = path;
    _sql_editors[editor_index].dirty = false;
    _sql_editors[editor_index].is_scratch = false;
    _sql_editors[editor_index].caption = g_basename(path.c_str());
  
    do_partial_ui_refresh(RefreshEditorTitle);
    
    _context_ui->get_wb()->add_recent_file(path);
    
    _grtm->replace_status_text(strfmt(_("SQL script saved to '%s'."), path.c_str()));
    return true;
  }
  return false;
}


std::string Db_sql_editor::restore_sql_from_history(int entry_index, std::list<int> &detail_indexes)
{
  return _history->restore_sql_from_history(entry_index, detail_indexes);
}


void Db_sql_editor::get_schemata(std::list<std::string> &schemata_names)
{
  try
  {
    GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();
    workbench_physical_ModelRef model= workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model());
    db_CatalogRef catalog= model->catalog();

    ListRef<db_Schema> schemata= catalog->schemata();
    if (0 == schemata.count())
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();

      bool show_metadata_schemata= (0 != _options.get_int("DbSqlEditor:ShowMetadataSchemata", 0));
      db_mgmt_RdbmsRef rdbms= db_mgmt_RdbmsRef::cast_from(grt::copy_object(_grtm->get_grt(), this->rdbms()));
      std::string database_package= *rdbms->databaseObjectPackage();

      std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemata());
      while (rs->next())
      {
        std::string name= rs->getString(1);
        db_SchemaRef schema= find_named_object_in_list(schemata, name);
        if (!schema.is_valid())
        {
          static std::map<std::string, bool> metadata_schemata_names;
          class MetadataSchemataNamesInitializer
          {
          public:
            MetadataSchemataNamesInitializer(std::map<std::string, bool> &metadata_schemata_names)
            {
              //! dbms-specific code
              metadata_schemata_names["information_schema"];
              metadata_schemata_names["performance_schema"];
              metadata_schemata_names["mysql"];
            }
          };
          static MetadataSchemataNamesInitializer metadata_schemata_initializer(metadata_schemata_names);

          if (show_metadata_schemata || (metadata_schemata_names.end() == metadata_schemata_names.find(name)))
          {
            schema= _grtm->get_grt()->create_object<db_Schema>(database_package + ".Schema");
            schema->owner(model->catalog());
            schema->modelOnly(1);
            schema->name(name);
            schemata.insert(schema);
          }
        }
      }
    }

    for (int n= 0, count= schemata.count(); n < count; ++n)
    {
      std::string name= *schemata.get(n)->name();
      schemata_names.push_back(name);
    }
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Get schemata"))
}


std::string Db_sql_editor::active_schema() const
{
  return (_usr_dbc_conn) ? _usr_dbc_conn->active_schema : std::string();
}


int Db_sql_editor::active_schema_index() const
{
  std::string schema_name= active_schema();
  if (schema_name.empty())
  {
    GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();
    workbench_physical_ModelRef model= workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model());
    db_CatalogRef catalog= model->catalog();
    if (catalog->schemata().count() > 0)
      return 0;
    else
      return -1;
  }
  int res= _schema_tree.get_index_of_schema(schema_name);
  return res;
}


void Db_sql_editor::cache_active_schema_name()
{
  std::string schema=_usr_dbc_conn->ref->getSchema();
  _usr_dbc_conn->active_schema= schema;
  _aux_dbc_conn->active_schema= schema;

  _schema_tree.set_active_schema(schema);
  if (_side_bar)
    _side_bar->get_schema_tree()->refresh(bec::NodeId());
}


void Db_sql_editor::active_schema(const std::string &value)
{
  try
  {
    if (value.empty())
      return;

    std::string schema= active_schema();
    if (value == schema)
      return;

    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      _aux_dbc_conn->ref->setSchema(value);
      _aux_dbc_conn->active_schema= value;
    }

    {
      GMutexLock usr_dbc_conn_mutex= ensure_valid_usr_connection();
      _usr_dbc_conn->ref->setSchema(value);
      _usr_dbc_conn->active_schema= value;
    }
    
    _schema_tree.set_active_schema(value);
    if (_side_bar)
      _side_bar->get_schema_tree()->refresh(bec::NodeId());

    grt_manager()->replace_status_text(strfmt(_("Active schema changed to %s"), value.c_str()));
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Set active schema"))
}


wb::LiveSchemaTree *Db_sql_editor::get_schema_tree()
{
  return &_schema_tree;
}


std::list<std::string> Db_sql_editor::fetch_live_schema_list()
{
  std::list<std::string> schemata;
  get_schemata(schemata);
  return schemata;
}


void Db_sql_editor::notify_of_fetched_schema_contents(
  Ptr self_ptr,
  bec::NodeId node,
  db_SchemaRef schema,
  bool schemata_tree_needs_refresh,
  bool live_physical_overview_schema_node_needs_refresh,
  bool live_physical_overview_needs_refresh)
{
  RETURN_IF_FAIL_TO_RETAIN_WEAK_PTR (Db_sql_editor, self_ptr, self)

  if (schemata_tree_needs_refresh)
  {
    _schema_tree.tree_changed(0, node);
//    _schema_tree.refresh_node_ui.emit(node);
  }

  if (live_physical_overview_schema_node_needs_refresh)
  {
    _live_physical_overview->refresh_schema_node(schema);
  }

  if (live_physical_overview_needs_refresh)
  {
#if defined(__APPLE__) || !defined(_WIN32)
    // all platforms should refresh on RefreshOverview or some variation of it
    // send_refresh_for_schema() is something that was meant for model overview, not live --alfredo
    do_partial_ui_refresh(RefreshOverview);
#else
    _live_physical_overview->send_refresh_for_schema(schema, false);
#endif
  }
}


bool Db_sql_editor::fetch_live_schema_contents(bec::NodeId node_id, const std::string &schema_name, const wb::LiveSchemaTree::SchemaContentArrivedSlot &arrived_slot)
{
  // in windows we use TreeViewAdv feature to expand nodes asynchronously
  // that is this function is already called from a separate thread
  // and it must have items loaded when it returns.
  bool sync= !_grtm->in_main_thread();

  live_schema_fetch_task->exec(sync,
    sigc::bind(sigc::mem_fun(this, &Db_sql_editor::do_fetch_live_schema_contents),
      weak_ptr_from(this), node_id, schema_name, arrived_slot));

  return true;//!
}


void Db_sql_editor::refresh_live_object_in_overview(wb::LiveSchemaTree::ObjectType type, const std::string schema_name, const std::string old_obj_name, const std::string new_obj_name)
{
  try
  {
    //GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection(); // this lock is acquired by caller
    GMutexLock schema_contents_mutex(_schema_contents_mutex);

    bool is_obj_created= old_obj_name.empty();
    bool is_obj_dropped= new_obj_name.empty();
    bool is_obj_renamed= !is_obj_created && !is_obj_dropped && (old_obj_name != new_obj_name);
    if (is_obj_renamed)
      is_obj_created= is_obj_dropped= true;

    // update schema tree even if no object was added/dropped, to clear details attribute which contents might to be changed
    _schema_tree.update_live_object_state(type, schema_name, old_obj_name, new_obj_name);

    if (!is_obj_created && !is_obj_dropped)
      return;

    std::string obj_type;
    switch (type)
    {
      case wb::LiveSchemaTree::Schema:
        obj_type= "schema";
        break;
      case wb::LiveSchemaTree::Table:
        obj_type= "table";
        break;
      case wb::LiveSchemaTree::View:
        obj_type= "view";
        break;
      case wb::LiveSchemaTree::Routine:
        obj_type= "routine";
        break;
    }

    if (is_obj_created)
    {
      is_obj_created= false;
      std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, obj_type, false, new_obj_name));
      while (rs->next())
        is_obj_created= true;
    }
    if (is_obj_dropped)
    {
      std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, obj_type, false, old_obj_name));
      while (rs->next())
        is_obj_dropped= false;
    }

    if (!is_obj_created && !is_obj_dropped)
      return;

    db_SchemaRef schema;

    {
      GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();

      workbench_physical_ModelRef model= workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model());
      db_CatalogRef catalog= model->catalog();
      db_mgmt_RdbmsRef rdbms= this->rdbms();
      std::string database_package= *rdbms->databaseObjectPackage();

      schema= find_named_object_in_list(catalog->schemata(), schema_name);

      switch (type)
      {
        case wb::LiveSchemaTree::Schema:
          {
            if (is_obj_created)
            {
              db_SchemaRef obj= find_named_object_in_list(catalog->schemata(), new_obj_name);
              if (!obj.is_valid())
              {
                schema= _grtm->get_grt()->create_object<db_Schema>(database_package + ".Schema");
                schema->owner(model->catalog());
                schema->modelOnly(1);
                schema->name(new_obj_name);
                catalog->schemata().insert(schema);
              }
            }
            if (is_obj_dropped)
            {
              db_SchemaRef obj= find_named_object_in_list(catalog->schemata(), old_obj_name);
              if (obj.is_valid())
                catalog->schemata().remove_value(obj);
            }
          }
          break;

        case wb::LiveSchemaTree::Table:
          {
            if (is_obj_created)
            {
              db_TableRef obj= find_named_object_in_list(schema->tables(), new_obj_name);
              if (!obj.is_valid())
              {
                db_TableRef table= _grtm->get_grt()->create_object<db_Table>(database_package + ".Table");
                table->owner(schema);
                table->name(new_obj_name);
                schema->tables().insert(table);
              }
            }
            if (is_obj_dropped)
            {
              db_TableRef obj= find_named_object_in_list(schema->tables(), old_obj_name);
              if (obj.is_valid())
                schema->tables().remove_value(obj);
            }
          }
          break;
        case wb::LiveSchemaTree::View:
          {
            if (is_obj_created)
            {
              db_ViewRef obj= find_named_object_in_list(schema->views(), new_obj_name);
              if (!obj.is_valid())
              {
                db_ViewRef view= _grtm->get_grt()->create_object<db_View>(database_package + ".View");
                view->owner(schema);
                view->name(new_obj_name);
                schema->views().insert(view);
              }
            }
            if (is_obj_dropped)
            {
              db_ViewRef obj= find_named_object_in_list(schema->views(), old_obj_name);
              if (obj.is_valid())
                schema->views().remove_value(obj);
            }
          }
          break;
        case wb::LiveSchemaTree::Routine:
          {
            if (is_obj_created)
            {
              db_RoutineRef obj= find_named_object_in_list(schema->routines(), new_obj_name);
              if (!obj.is_valid())
              {
                db_RoutineRef routine= _grtm->get_grt()->create_object<db_Routine>(database_package + ".Routine");
                routine->owner(schema);
                routine->name(new_obj_name);
                schema->routines().insert(routine);
              }
            }
            if (is_obj_dropped)
            {
              db_RoutineRef obj= find_named_object_in_list(schema->routines(), old_obj_name);
              if (obj.is_valid())
                schema->routines().remove_value(obj);
            }
          }
          break;
      }
    }

    if (wb::LiveSchemaTree::Schema == type)
    {
      do_refresh_physical_overview(false);
    }
    else
    {
      NodeId schema_node_id= _schema_tree.get_node_for_schema(schema_name);
      live_schema_fetch_task->execute_in_main_thread(
        sigc::bind(sigc::mem_fun(this, &Db_sql_editor::notify_of_fetched_schema_contents),
          weak_ptr_from(this), schema_node_id, schema, true, true, true),
        true,
        false);
    }
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH_TO_DEFAULT_LOG(_("Refresh live schema object"))
}


grt::StringRef Db_sql_editor::do_fetch_live_schema_contents(grt::GRT *grt, Ptr self_ptr, bec::NodeId node_id, std::string schema_name, wb::LiveSchemaTree::SchemaContentArrivedSlot arrived_slot)
{
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Db_sql_editor, self_ptr, self, grt::StringRef(""))

  try
  {
    GMutexLock schema_contents_mutex(_schema_contents_mutex);
    GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();

    workbench_physical_ModelRef model= workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model());
    db_CatalogRef catalog= model->catalog();
    db_SchemaRef schema= find_named_object_in_list(catalog->schemata(), schema_name);

    if (!schema.is_valid())
      return grt::StringRef("");

    if (!schema->modelOnly() && arrived_slot.empty())
      return grt::StringRef("");

    bool live_physical_overview_schema_node_needs_refresh= false;
#ifdef _WIN32
    bool live_physical_overview_needs_refresh= schema->modelOnly() && arrived_slot.empty();
#else
    bool live_physical_overview_needs_refresh= true;
#endif

    LiveSchemaTree::SchemaNode node;
    node.fetched= true;
    node.name= schema_name;

    if (schema->modelOnly())
    {
      schema->modelOnly(0);

      db_mgmt_RdbmsRef rdbms= this->rdbms();
      std::string database_package= *rdbms->databaseObjectPackage();

      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();

      {
        std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "table", false));
        while (rs->next())
        {
          wb::LiveSchemaTree::ObjectNode obj_node;
          obj_node.name= rs->getString(4);
          node.tables.push_back(obj_node);

          db_TableRef table= _grtm->get_grt()->create_object<db_Table>(database_package + ".Table");
          table->owner(schema);
          table->name(obj_node.name);
          schema->tables().insert(table);
        }
      }
      
      {
        std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "view", false));
        while (rs->next())
        {
          wb::LiveSchemaTree::ObjectNode obj_node;
          obj_node.name= rs->getString(4);
          node.views.push_back(obj_node);

          db_ViewRef view= _grtm->get_grt()->create_object<db_View>(database_package + ".View");
          view->owner(schema);
          view->name(obj_node.name);
          schema->views().insert(view);
        }
      }
      
      {
        std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "routine", false));
        while (rs->next())
        {
          wb::LiveSchemaTree::ObjectNode obj_node;
          obj_node.name= rs->getString(4);
          obj_node.details= " "; // means no details
          node.routines.push_back(obj_node);

          db_RoutineRef routine= _grtm->get_grt()->create_object<db_Routine>(database_package + ".Routine");
          routine->owner(schema);
          routine->name(obj_node.name);
          schema->routines().insert(routine);
        }
      }

      live_physical_overview_schema_node_needs_refresh= true;
    }
    else
    {
      for (size_t n= 0, count= schema->tables().count(); n < count; ++n)
      {
        wb::LiveSchemaTree::ObjectNode obj_node;
        obj_node.name= *schema->tables().get(n)->name();
        node.tables.push_back(obj_node);
      }
      for (size_t n= 0, count= schema->views().count(); n < count; ++n)
      {
        wb::LiveSchemaTree::ObjectNode obj_node;
        obj_node.name= *schema->views().get(n)->name();
        node.views.push_back(obj_node);
      }
      for (size_t n= 0, count= schema->routines().count(); n < count; ++n)
      {
        wb::LiveSchemaTree::ObjectNode obj_node;
        obj_node.name= *schema->routines().get(n)->name();
        obj_node.details= " "; // means no details
        node.routines.push_back(obj_node);
      }
    }

    arrived_slot(node);

    live_schema_fetch_task->execute_in_main_thread(
      sigc::bind(sigc::mem_fun(this, &Db_sql_editor::notify_of_fetched_schema_contents),
        weak_ptr_from(this), node_id, schema, !arrived_slot.empty(),
        live_physical_overview_schema_node_needs_refresh, live_physical_overview_needs_refresh),
      false,
      false);
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH_TO_DEFAULT_LOG(_("Get schema contents"))

  return grt::StringRef("");
}


bool Db_sql_editor::fetch_object_details(wb::LiveSchemaTree::ObjectType obj_type, const std::string &schema_name, const std::string &obj_name, const wb::LiveSchemaTree::ObjectDetailsArrivedSlot &arrived_slot)
{
  if ((obj_type == wb::LiveSchemaTree::Table) || (obj_type == wb::LiveSchemaTree::View))
  {
    LiveSchemaTree::SchemaNode schema_node;
    schema_node.fetched= true;
    schema_node.name= schema_name;

    LiveSchemaTree::ObjectNode obj_node;
    obj_node.fetched= true;
    obj_node.name= obj_name;

    std::string col_brief_list;
    std::string col_detail_list;
    /*
    try
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getColumns("", schema_name, obj_name, "%"));
      while (rs->next())
      {
        std::string column_name= rs->getString(4);
        obj_node.column_names.push_back(column_name);
        col_brief_list+= (col_brief_list.empty() ? "" : ", ") + column_name;
        col_detail_list+= strfmt("%-12s", column_name.c_str()) + " " + rs->getString(6) + "\n";
      }
    }
    catch (const sql::SQLException&)
    {
    }*/
    
    // get key info (we could use getExportedKeys() to get PK and FK info, but since it relies on very slow I_S, better not for now)
    try
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      
      std::auto_ptr<sql::Statement> stmt(_aux_dbc_conn->ref->createStatement());
      std::auto_ptr<sql::ResultSet> rs(stmt->executeQuery(strfmt("SHOW COLUMNS FROM `%s`.`%s`", schema_name.c_str(), obj_name.c_str())));

      while (rs->next())
      {
        std::string column_name= rs->getString(1);
        std::string type = rs->getString(2);
        // 3 Null
        std::string key = rs->getString(4);
        // 5 Default
        // 6 Extra
        
        LiveSchemaTree::ColumnNode col_node;
        col_node.name = column_name;
        col_node.is_pk = key == "PRI";

        obj_node.columns.push_back(col_node);
        col_brief_list+= (col_brief_list.empty() ? "" : ", ") + column_name;
        if (key == "PRI")
          col_detail_list+= strfmt("%-16s", column_name.c_str()) + " " + type + " PK\n";
        else
          col_detail_list+= strfmt("%-16s", column_name.c_str()) + " " + type + "\n";
      }
    }
    catch (const sql::SQLException& exc)
    {
      g_warning("Error fetching column information: %s", exc.what());
    }

    obj_node.details= strfmt(_("Table %s\n"), obj_name.c_str());
    size_t header_length= obj_node.details.size()-1;
    obj_node.details+= std::string(header_length, '=') + "\n" +
      col_brief_list + "\n" +
      std::string(header_length, '-').append("\n") +
      col_detail_list;

    arrived_slot(schema_node, obj_node, obj_type);

    return true;
  }
  return false;
}


void Db_sql_editor::schema_object_activated(const std::string &action, wb::LiveSchemaTree::ObjectType type, const std::string &schema, const std::string &name)
{
  std::string action_modifier; // action can contain prefix denoting action modifier
  std::string real_action= action; // if action modifier is present real_action will store part after w/o modifier prefix
  const char *action_modifier_clipboard= "clipboard_"; // denotes copy of the action result to clipboard
  const char *action_modifier_editor= "editor_"; // denotes insertion of the action result into editor at cursor position
  {
    const char* known_modifiers[]= { action_modifier_clipboard, action_modifier_editor };
    for (size_t n= 0, size= sizeof(known_modifiers)/sizeof(known_modifiers[0]); n < size; ++n)
    {
      if (0 == action.find(known_modifiers[n]))
      {
        action_modifier= known_modifiers[n];
        real_action= action.substr(action_modifier.size());
      }
    }
  }

  if (real_action == "object_name_short" || real_action == "object_name_long")
  {
    std::string text;
    if (type == wb::LiveSchemaTree::Schema)
      text= name;
    else if (real_action == "object_name_short")
      text= name;
    else if (real_action == "object_name_long")
      text= strfmt("`%s`.`%s`", schema.c_str(), name.c_str());
    if (action_modifier == action_modifier_clipboard)
      mforms::Utilities::set_clipboard_text(text);
    else if (action_modifier == action_modifier_editor)
      sql_editor()->insert_text(text);
    return;
  }
  else if (real_action == "table_create_statement")
  {
    std::string text= get_object_ddl_script(type, schema, name);
    if (action_modifier == action_modifier_clipboard)
      mforms::Utilities::set_clipboard_text(text);
    else if (action_modifier == action_modifier_editor)
      sql_editor()->insert_text(text);
  }
  else if (real_action == "table_column_info_statement")
  {
    std::string text= strfmt("SHOW COLUMNS FROM `%s`.`%s`;\n", schema.c_str(), name.c_str());
    if (action_modifier == action_modifier_clipboard)
      mforms::Utilities::set_clipboard_text(text);
    else if (action_modifier == action_modifier_editor)
      sql_editor()->insert_text(text);
  }
  else if (real_action == "table_delete_statement")
  {
    std::string text= strfmt("DELETE FROM `%s`.`%s`\nWHERE <where_condition>;\n", schema.c_str(), name.c_str());
    if (action_modifier == action_modifier_clipboard)
      mforms::Utilities::set_clipboard_text(text);
    else if (action_modifier == action_modifier_editor)
      sql_editor()->insert_text(text);
    return;
  }
  else if (real_action == "table_select_statement" || real_action == "table_insert_statement" ||
    real_action == "table_update_statement" || real_action == "table_column_names")
  {
    typedef std::list<std::pair<std::string, std::string> > ColumnDefinitions;
    ColumnDefinitions column_defs;
    try
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getColumns("", schema, name, "%"));
      while (rs->next())
        column_defs.push_back(std::make_pair(rs->getString(4), rs->getString(6)));
    }
    catch (const sql::SQLException&)
    {
    }

    std::string text;

    if (real_action == "table_select_statement")
    {
      std::string columns;
      BOOST_FOREACH (const ColumnDefinitions::value_type &col_def, column_defs)
        columns+= (columns.empty() ? "" : ",\n") + strfmt("`%s`.`%s`", name.c_str(), col_def.first.c_str());
      text= strfmt("SELECT\n%s\nFROM `%s`.`%s`;\n", columns.c_str(), schema.c_str(), name.c_str());
    }

    if (real_action == "table_insert_statement")
    {
      std::string columns;
      std::string values;
      BOOST_FOREACH (const ColumnDefinitions::value_type &col_def, column_defs)
      {
        columns+= (columns.empty() ? "" : ",\n") + strfmt("`%s`", col_def.first.c_str());
        values+= (values.empty() ? "" : ",\n") + strfmt("{%s: %s}", col_def.first.c_str(), col_def.second.c_str());
      }
      text= strfmt("INSERT INTO `%s`.`%s`\n(%s)\nVALUES\n(\n%s\n);\n", schema.c_str(), name.c_str(), columns.c_str(), values.c_str());
    }

    if (real_action == "table_update_statement")
    {
      std::string columns;
      BOOST_FOREACH (const ColumnDefinitions::value_type &col_def, column_defs)
        columns+= (columns.empty() ? "" : ",\n") + strfmt("`%s` = {%s: %s}", col_def.first.c_str(), col_def.first.c_str(), col_def.second.c_str());
      text= strfmt("UPDATE `%s`.`%s`\nSET\n%s\nWHERE <where_condition>;\n", schema.c_str(), name.c_str(), columns.c_str());
    }

    if (real_action == "table_column_names")
    {
      BOOST_FOREACH (const ColumnDefinitions::value_type &col_def, column_defs)
        text+= (text.empty() ? "" : ", ") + strfmt("`%s`.`%s`", name.c_str(), col_def.first.c_str());
    }

    if (action_modifier == action_modifier_clipboard)
      mforms::Utilities::set_clipboard_text(text);
    else if (action_modifier == action_modifier_editor)
      sql_editor()->insert_text(text);

    return;
  }
  
  {
    std::string sql;
    switch (type)
    {
    case wb::LiveSchemaTree::Schema:
      active_schema(schema);
      break;
    case wb::LiveSchemaTree::Table:
      if (real_action == "activate" || real_action == "edit_data")
        sql= strfmt("EDIT `%s`.`%s`;", schema.c_str(), name.c_str());
      else if (real_action == "select_data")
        sql= strfmt("SELECT * FROM `%s`.`%s`;", schema.c_str(), name.c_str());
      break;
    case wb::LiveSchemaTree::View:
      if (real_action == "activate" || real_action == "select_data")
        sql= strfmt("SELECT * FROM `%s`.`%s`;", schema.c_str(), name.c_str());
      break;
    default:
      break;
    }
    if (!sql.empty())
    {      
      do_partial_ui_refresh(Db_sql_editor::RefreshEditorBackend);
      Sql_editor::Ref editor(sql_editor());

      if (!editor)
        new_sql_scratch_area();
      else
      {
        if (!_sql_editors[_active_sql_editor_index].is_scratch)
          new_sql_scratch_area();
      }
      
      set_sql_editor_text(sql, true);
      sql_editor_dirty(_active_sql_editor_index, false);
    }
  }
}


void Db_sql_editor::do_alter_live_object(wb::LiveSchemaTree::ObjectType type, const std::string &schema_name, const std::string &aobj_name)
{
  std::string obj_name = aobj_name;
  try
  {
    db_mgmt_RdbmsRef rdbms= this->rdbms();
    std::string database_package= *rdbms->databaseObjectPackage();

    db_CatalogRef client_state_catalog= _grtm->get_grt()->create_object<db_Catalog>(database_package + ".Catalog");
    client_state_catalog->name("default");
    client_state_catalog->oldName("default");
    client_state_catalog->version(rdbms->version());
    grt::replace_contents(client_state_catalog->simpleDatatypes(), rdbms->simpleDatatypes());
    grt::replace_contents(client_state_catalog->userDatatypes(), 
                          workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model())->catalog()->userDatatypes());

    db_SchemaRef schema;
    if (wb::LiveSchemaTree::Schema != type)
    {
      schema= _grtm->get_grt()->create_object<db_Schema>(database_package + ".Schema");
      schema->owner(client_state_catalog);
      schema->name(schema_name);
      schema->oldName(schema_name);
      client_state_catalog->schemata().insert(schema);
      client_state_catalog->defaultSchema(schema);
    }

    bool is_object_new= obj_name.empty();

    std::string ddl_script;
    std::string sql_mode;
    if (!is_object_new)
    {
      // parse selected object DDL into auxiliary catalog
      ddl_script= get_object_ddl_script(type, schema_name, obj_name);
      if (ddl_script.empty())
        return;
      SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms);
      Sql_parser::Ref sql_parser= sql_facade->sqlParser();
      sql_parser->messages_enabled(false);
      grt::DictRef options(_grtm->get_grt());
      {
        GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
        if (get_session_variable(_aux_dbc_conn, "sql_mode", sql_mode))
          options.gset("sql_mode", sql_mode);
      }
      int err_count= sql_parser->parse_sql_script(client_state_catalog, ddl_script, options);

      //! dbms-specific code
      if (options.has_key("sql_mode") && (0 < err_count))
      {
        if (std::string::npos != sql_mode.find("ANSI_QUOTES"))
          sql_mode= replace_string(sql_mode, "ANSI_QUOTES", "");
        else
          sql_mode+= ",ANSI_QUOTES";
        options.gset("sql_mode", sql_mode);
        options.set("reuse_existing_objects", grt::IntegerRef(1));
        err_count= sql_parser->parse_sql_script(client_state_catalog, ddl_script, options);
        if (0 == err_count)
        {
          mforms::Utilities::show_warning(strfmt(_("Problem with parsing object's DDL: `%s`.`%s`"), schema_name.c_str(), obj_name.c_str()),
            _("The object's DDL retrieved from DBMS is inconsistent with SQL_MODE variable set for the connection. "
              "In particular current state of ANSI_QUOTES flag contradicts to what it was set when the object was being created. "
              "This may lead to fail when trying to apply changes. "
              "As a workaround you might want temporarily to change the SQL_MODE variable to its previous value."),
            _("OK"));
        }
      }
    }

    db_CatalogRef server_state_catalog(db_CatalogRef::cast_from(grt::copy_object(_grtm->get_grt(), client_state_catalog)));

  retry_search:
    db_DatabaseObjectRef db_object;
    switch (type)
    {
      case wb::LiveSchemaTree::Schema:
        db_object= is_object_new ?
          create_new_schema(client_state_catalog) :
          find_named_object_in_list(client_state_catalog->schemata(), obj_name);
        break;
      case wb::LiveSchemaTree::Table:
        db_object= is_object_new ?
          create_new_table(schema) :
          find_named_object_in_list(schema->tables(), obj_name);
        break;
      case wb::LiveSchemaTree::View:
        db_object= is_object_new ?
          create_new_view(schema) :
          find_named_object_in_list(schema->views(), obj_name);
        break;
      case wb::LiveSchemaTree::Routine:
        db_object= is_object_new ?
          create_new_routine(schema) :
          find_named_object_in_list(schema->routines(), obj_name);
        break;
    }
    if (!db_object.is_valid())
    {
      std::string lower;
      // There is a bug in server that causes get_object_ddl_script() for uppercase named tables
      // to be returned in lowercase, in osx. http://bugs.mysql.com/bug.php?id=57830
      // So, if you try to alter FOOBAR, it will not find it, since the table will be revenged
      // in lowercase. To work around that, we convert the object name to lowercase and repeat
      // the search if the 1st try didn't work.
      lower = obj_name;
      base::tolower(lower);
      if (lower != obj_name)
      {
        g_warning("Object name %s was not found in catalog, trying to search it as %s",
                  obj_name.c_str(), lower.c_str());
        obj_name = lower;
        goto retry_search;
      }
      else
        g_warning("Object name %s was not found in catalog.", aobj_name.c_str());

      return;
    }
    db_object->customData().set("liveRdbms", rdbms);
    db_object->customData().set("clientStateCatalog", client_state_catalog);
    db_object->customData().set("serverStateCatalog", server_state_catalog);
    db_object->customData().set("originalObjectDDL", grt::StringRef(ddl_script));
    db_object->customData().set("sqlMode", grt::StringRef(sql_mode));
    {
      std::ostringstream oss;
      oss << this;
      db_object->customData().set("contextDbSqlEditor", grt::StringRef(oss.str()));
    }

    _context_ui->get_wb()->open_object_editor(db_object, bec::StandaloneWindowFlag);
  }
  catch (const std::exception & e)
  {
    mforms::Utilities::show_error(strfmt(_("Failed to create/alter `%s`.`%s`"), schema_name.c_str(), obj_name.c_str()), e.what(), _("OK"));
  }
}


db_SchemaRef Db_sql_editor::create_new_schema(db_CatalogRef owner)
{
  db_SchemaRef object= _grtm->get_grt()->create_object<db_Schema>(owner->schemata()->content_type_spec().object_class);
  object->owner(owner);
  object->name("new_schema");
  owner->schemata().insert(object);
  owner->defaultSchema(object);
  return object;
}


db_TableRef Db_sql_editor::create_new_table(db_SchemaRef owner)
{
  db_TableRef object= _grtm->get_grt()->create_object<db_Table>(owner->tables()->content_type_spec().object_class);
  object->owner(owner);
  object->name("new_table");
  owner->tables().insert(object);
  return object;
}


db_ViewRef Db_sql_editor::create_new_view(db_SchemaRef owner)
{
  db_ViewRef object= _grtm->get_grt()->create_object<db_View>(owner->views()->content_type_spec().object_class);
  object->owner(owner);
  object->name("new_view");
  owner->views().insert(object);
  return object;
}


db_RoutineRef Db_sql_editor::create_new_routine(db_SchemaRef owner)
{
  db_RoutineRef object= _grtm->get_grt()->create_object<db_Routine>(owner->routines()->content_type_spec().object_class);
  object->owner(owner);
  object->name("new_routine");
  owner->routines().insert(object);
  return object;
}


void Db_sql_editor::do_create_live_object(wb::LiveSchemaTree::ObjectType type, const std::string &schema_name, const std::string &obj_name)
{
  do_alter_live_object(type, schema_name, obj_name);
}


void Db_sql_editor::do_drop_live_object(wb::LiveSchemaTree::ObjectType type, const std::string schema_name, const std::string obj_name)
{
  std::string full_obj_identifier;
  std::string obj_type;

  try
  {
    std::string sql;
    std::string title;
    switch (type)
    {
      case wb::LiveSchemaTree::Schema:
        full_obj_identifier= strfmt("`%s`", obj_name.c_str());
        obj_type= "schema";
        title = "DROP DATABASE";
        break;
      case wb::LiveSchemaTree::Table:
        full_obj_identifier= strfmt("`%s`.`%s`", schema_name.c_str(), obj_name.c_str());
        obj_type= "table";
        title = "DROP TABLE";
        break;
      case wb::LiveSchemaTree::View:
        full_obj_identifier= strfmt("`%s`.`%s`", schema_name.c_str(), obj_name.c_str());
        obj_type= "view";
        title = "DROP VIEW";
        break;
      case wb::LiveSchemaTree::Routine:
        {
          int routine_type= 0;
          title = "DROP ROUTINE";
          {
            std::string query_sql= strfmt("select routine_type from information_schema.routines where routine_schema='%s' and routine_name='%s'", schema_name.c_str(), obj_name.c_str());
            {
              GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
              std::auto_ptr<sql::Statement> stmt(_aux_dbc_conn->ref->createStatement());
              boost::shared_ptr<sql::ResultSet> rs(stmt->executeQuery(query_sql));
              if (rs->next())
                routine_type= (rs->getString(1) == "PROCEDURE") ? 1 : -1;
            }
          }
          if (routine_type)
          {
            full_obj_identifier= strfmt("`%s`.`%s`", schema_name.c_str(), obj_name.c_str());
            obj_type= (routine_type == 1) ? "procedure" : "function";
            title = (routine_type == 1) ? "DROP PROCEDURE" : "DROP FUNCTION";
          }
        }
        break;
    }

    if (!obj_type.empty())
    {
      sql= strfmt("drop %s %s", obj_type.c_str(), full_obj_identifier.c_str());
      if (run_live_object_alteration_wizard(sql, NULL))
      {
        refresh_live_object_in_overview(type, schema_name, obj_name, "");
        if ((wb::LiveSchemaTree::Schema == type) && (active_schema() == obj_name) && _connection.is_valid())
        {
          std::string default_schema= _connection->parameterValues().get_string("schema", "");
          if (!default_schema.empty())
          {
            _grtm->get_dispatcher()->call_from_main_thread<void>(
              sigc::bind(sigc::mem_fun(this, (void(Db_sql_editor::*)(const std::string&))&Db_sql_editor::active_schema), default_schema),
              false/*wait*/,
              true/*force_queue*/);
          }
        }
      }
    }
  }
  catch (const std::exception &e)
  {
    mforms::Utilities::show_error(strfmt(_("Failed to drop %s %s"), obj_type.c_str(), full_obj_identifier.c_str()), e.what(), _("OK"));
  }
}


std::string Db_sql_editor::get_object_ddl_script(wb::LiveSchemaTree::ObjectType type, const std::string &schema_name, const std::string &obj_name)
{
  const size_t DDL_COLUMN= 5;
  std::string delimiter;
  {
    SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(rdbms());
    Sql_specifics::Ref sql_specifics= sql_facade->sqlSpecifics();
    delimiter= sql_specifics->non_std_sql_delimiter();
  }
  std::string ddl_script;
  ddl_script+= strfmt("delimiter %s\n\n", delimiter.c_str());

  try
  {
    GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();

    std::auto_ptr<sql::ResultSet> ddl_rs;

    switch (type)
    {
      case wb::LiveSchemaTree::Schema:
        ddl_rs.reset(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", obj_name, "schema", true, obj_name));
        break;

      case wb::LiveSchemaTree::Table:
        {
          // triggers
          std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "trigger", true, "", obj_name));
          while (rs->next())
          {
            ddl_script+= rs->getString(DDL_COLUMN);
            ddl_script+= delimiter + "\n\n";
          }
        }
        ddl_rs.reset(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "table", true, obj_name));
        break;

      case wb::LiveSchemaTree::View:
        ddl_rs.reset(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "view", true, obj_name));
        break;

      case wb::LiveSchemaTree::Routine:
        ddl_rs.reset(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "routine", true, obj_name));
        break;
    }

    while (ddl_rs->next())
    {
      ddl_script+= ddl_rs->getString(DDL_COLUMN);
      ddl_script+= delimiter + "\n\n";
    }
  }
  catch (const sql::SQLException &)
  {
    ddl_script.clear();
  }
  return ddl_script;
}


void Db_sql_editor::refresh_live_object_in_editor(bec::DBObjectEditorBE* obj_editor, bool using_old_name)
{
  db_DatabaseObjectRef db_object= obj_editor->get_dbobject();

  db_CatalogRef client_state_catalog= db_CatalogRef::cast_from(db_object->customData().get("clientStateCatalog"));

  std::string obj_name= using_old_name ? db_object->oldName() : db_object->name();
  // don't refresh new objects that where not applied yet
  if (obj_name.empty())
    return;
  db_object->name(obj_name);
  db_object->oldName("");

  std::string schema_name= db_SchemaRef::can_wrap(db_object) ? std::string() : *db_object->owner()->name();
  db_SchemaRef schema;
  if (!schema_name.empty())
    schema= db_SchemaRef::cast_from(db_object->owner());

  wb::LiveSchemaTree::ObjectType db_object_type;

  if (db_SchemaRef::can_wrap(db_object))
  {
    db_object_type= wb::LiveSchemaTree::Schema;
  }
  else
  {
    if (db_TableRef::can_wrap(db_object))
    {
      db_object_type= wb::LiveSchemaTree::Table;

      // reset selection of fkeys/indices to avoid warnings
      bec::TableEditorBE *table_editor= dynamic_cast <bec::TableEditorBE *> (obj_editor);
      table_editor->get_fks()->select_fk(NodeId());
      table_editor->get_indexes()->select_index(NodeId());

      db_TableRef table= db_TableRef::cast_from(db_object);
      table->isStub(1);
      table->triggers().remove_all();
      table->foreignKeys().remove_all();
      table->indices().remove_all();
      table->columns().remove_all();
    }
    else if (db_ViewRef::can_wrap(db_object))
    {
      db_object_type= wb::LiveSchemaTree::View;
    }
    else if (db_RoutineRef::can_wrap(db_object))
    {
      db_object_type= wb::LiveSchemaTree::Routine;
    }
  }

  // reparse object's DDL
  std::string ddl_script;
  std::string sql_mode;
  {
    ddl_script= get_object_ddl_script(db_object_type, schema_name, obj_name);
    if (!ddl_script.empty())
    {
      db_object->oldName(obj_name);

      SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(obj_editor->get_rdbms());
      Sql_parser::Ref sql_parser= sql_facade->sqlParser();
      sql_parser->messages_enabled(false);
      grt::DictRef options(_grtm->get_grt());
      options.set("reuse_existing_objects", grt::IntegerRef(1));
      try
      {
        GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
        if (get_session_variable(_aux_dbc_conn, "sql_mode", sql_mode))
          options.gset("sql_mode", sql_mode);
      }
      CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Get 'sql_mode' session variable"));
      int err_count= sql_parser->parse_sql_script(client_state_catalog, ddl_script, options);

      //! dbms-specific code
      if (options.has_key("sql_mode") && (0 < err_count))
      {
        if (std::string::npos != sql_mode.find("ANSI_QUOTES"))
          sql_mode= replace_string(sql_mode, "ANSI_QUOTES", "");
        else
          sql_mode+= ",ANSI_QUOTES";
        options.gset("sql_mode", sql_mode);
        options.set("reuse_existing_objects", grt::IntegerRef(1));
        err_count= sql_parser->parse_sql_script(client_state_catalog, ddl_script, options);
        if (0 == err_count)
        {
          mforms::Utilities::show_warning(strfmt(_("Problem with parsing object's DDL: `%s`.`%s`"), schema_name.c_str(), obj_name.c_str()),
            _("The object's DDL retrieved from DBMS is inconsistent with SQL_MODE variable set for the connection. "
              "In particular current state of ANSI_QUOTES flag contradicts to what it was set when the object was being created. "
              "This may lead to fail when trying to apply changes. "
              "As a workaround you might want temporarily to change the SQL_MODE variable to its previous value."),
            _("OK"));
        }
      }
    }
  }

  {
    db_CatalogRef server_state_catalog(db_CatalogRef::cast_from(grt::copy_object(_grtm->get_grt(), client_state_catalog)));
    db_object->customData().set("serverStateCatalog", server_state_catalog);
  }
  db_object->customData().set("originalObjectDDL", grt::StringRef(ddl_script));
  db_object->customData().set("sqlMode", grt::StringRef(sql_mode));

  // enable refresh of sql editor contents
  Sql_editor::Ref sql_editor= obj_editor->get_sql_editor();
  if (sql_editor)
  {
    sql_editor->is_refresh_enabled(true);
    sql_editor->is_sql_check_enabled(true);
    // provoke database object to refresh FE control contents
    db_object->signal_changed().emit("", grt::ValueRef());
  }
}


void Db_sql_editor::notify_of_ui_form_creation(bec::UIForm *form)
{
  DBObjectEditorBE *editor= dynamic_cast <DBObjectEditorBE *> (form);
  if (editor)
  {
    editor->on_apply_changes_to_live_object= sigc::mem_fun(this, &Db_sql_editor::apply_changes_to_object);
    editor->on_refresh_live_object= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::refresh_live_object_in_editor), true);
    editor->on_create_live_table_stubs= sigc::mem_fun(this, &Db_sql_editor::create_live_table_stubs);
    editor->on_expand_live_table_stub= sigc::mem_fun(this, &Db_sql_editor::expand_live_table_stub);
  }
}


template <typename T>
static bool remove_model_only_objects(grt::ListRef<T> obj_list)
{
  for (int n= obj_list.count()-1;  n >= 0; --n)
    if (obj_list.get(n)->modelOnly())
      obj_list.remove(n);
  return true;
}

static bool remove_model_only_objects(db_TableRef table)
{
  //remove_model_only_objects(table->columns()); // modelOnly is not a member of db_Column
  remove_model_only_objects(table->triggers());
  //remove_model_only_objects(table->indices()); // modelOnly is not a member of db_Index
  remove_model_only_objects(table->foreignKeys());
  return true;
}

static bool remove_model_only_objects(db_SchemaRef schema)
{
  remove_model_only_objects(schema->tables());
  schema->tables().foreach(sigc::ptr_fun((bool(*)(db_TableRef))&remove_model_only_objects));
  remove_model_only_objects(schema->views());
  remove_model_only_objects(schema->routines());
  return true;
}

static bool remove_model_only_objects(db_CatalogRef catalog)
{
  remove_model_only_objects(catalog->schemata());
  catalog->schemata().foreach(sigc::ptr_fun((bool(*)(db_SchemaRef))&remove_model_only_objects));
  return true;
}

bool Db_sql_editor::apply_changes_to_object(bec::DBObjectEditorBE* obj_editor, bool dry_run)
{
  try
  {
    if (!dry_run && obj_editor->get_sql_editor()->has_sql_errors())
    {
      int res= mforms::Utilities::show_warning(
        _("Apply Changes to Object"),
        _("The object's DDL statement contains syntax errors.\n"
          "Are you sure you want to apply the DDL statement unchanged?"),
        _("Yes"),
        _("No"));

      if (res != mforms::ResultOk)
        return false;
    }

    db_DatabaseObjectRef db_object= obj_editor->get_dbobject();

    // check for name conflicts
    // if the object is new or its name was changed
    if (!dry_run)
    {
      bool case_sensitive_identifiers= (_options.get_int("SqlIdentifiersCS", 1) != 0);
      std::string obj_name= db_object->name();
      std::string obj_old_name= db_object->oldName();
      if (!case_sensitive_identifiers)
      {
        base::tolower(obj_name);
        base::tolower(obj_old_name);
      }
      if (obj_name != obj_old_name)
      {
        std::list<std::string> obj_types;
        {
          if (db_SchemaRef::can_wrap(db_object))
          {
            obj_types.push_back("schema");
          }
          else if (db_TableRef::can_wrap(db_object) || db_ViewRef::can_wrap(db_object))
          {
            obj_types.push_back("table");
            obj_types.push_back("view");
          }
          else if (db_RoutineRef::can_wrap(db_object))
          {
            obj_types.push_back("routine");
          }
        }
        std::string schema_name= db_SchemaRef::can_wrap(db_object) ? std::string() : *db_object->owner()->name();
        BOOST_FOREACH (const std::string &obj_type, obj_types)
        {
          std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, obj_type, false, *db_object->name()));
          if (rs->next())
          {
            mforms::Utilities::show_error(
              _("Apply Changes to Object"),
              strfmt(_("Selected name conflicts with existing %s `%s`."), obj_type.c_str(), (*db_object->name()).c_str()),
              _("OK"));
            return false;
          }
        }
      }
    }

    DbMySQLImpl *diffsql_module= _grtm->get_grt()->find_native_module<DbMySQLImpl>("DbMySQL"); //! should be db-agnostic

    db_CatalogRef server_cat= db_CatalogRef::cast_from(db_object->customData().get("serverStateCatalog"));
    db_CatalogRef client_cat= db_CatalogRef::cast_from(db_object->customData().get("clientStateCatalog"));
    db_mgmt_RdbmsRef rdbms= db_mgmt_RdbmsRef::cast_from(db_object->customData().get("liveRdbms"));

    db_CatalogRef client_cat_copy= db_CatalogRef::cast_from(grt::copy_object(_grtm->get_grt(), client_cat));
    db_CatalogRef server_cat_copy= db_CatalogRef::cast_from(grt::copy_object(_grtm->get_grt(), server_cat));

    remove_model_only_objects(client_cat_copy);
    remove_model_only_objects(server_cat_copy);

    std::string alter_script= diffsql_module->makeAlterScriptForObject(server_cat_copy, client_cat_copy, db_object);

    if (alter_script.empty())
    {
      if (!dry_run)
        obj_editor->on_live_object_change_error(-1, _("No changes to object were detected."), "");
      return false; // no changes detected
    }

    if (dry_run)
      return true; // some changes were detected

    //bool is_live_object_alteration_wizard_enabled= (0 != _options.get_int("DbSqlEditor:IsLiveObjectAlterationWizardEnabled", 1));
    if (true/*is_live_object_alteration_wizard_enabled*/)
    {
      run_live_object_alteration_wizard(alter_script, obj_editor);
    }
    else
    {
      apply_object_alter_script(alter_script, obj_editor);
    }
  }
  catch (const std::exception &e)
  {
    obj_editor->on_live_object_change_error(-1, e.what(), "");
  }
  return true; // some changes were detected and applied
}


bool Db_sql_editor::run_live_object_alteration_wizard(std::string &alter_script, bec::DBObjectEditorBE* obj_editor)
{
  on_sql_script_run_error.clear();
  on_sql_script_run_progress.clear();
  on_sql_script_run_statistics.clear();

  SqlScriptRunWizard wizard(_grtm);
  on_sql_script_run_error.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_error));
  on_sql_script_run_progress.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_exec_progress));
  on_sql_script_run_statistics.connect(sigc::mem_fun(wizard.apply_page, &SqlScriptApplyPage::on_exec_stat));
  wizard.values().gset("sql_script", alter_script);
  wizard.apply_page->apply_sql_script= sigc::bind(sigc::mem_fun(this, &Db_sql_editor::apply_object_alter_script), obj_editor);
  wizard.run_modal();

  return wizard.applied() && !wizard.has_errors();
}


void Db_sql_editor::apply_object_alter_script(std::string &alter_script, bec::DBObjectEditorBE* obj_editor)
{
  SqlFacade::Ref sql_splitter= SqlFacade::instance_for_rdbms(rdbms());
  std::list<std::string> statements;
  sql_splitter->splitSqlScript(alter_script, statements);

  std::list<std::string> failback_statements;
  if (obj_editor)
  {
    // in case of alter script failure:
    // try to restore object since it could had been successfully dropped before the alter script failed
    db_DatabaseObjectRef db_object= obj_editor->get_dbobject();
    std::string original_object_ddl_script= grt::StringRef::cast_from(db_object->customData().get("originalObjectDDL"));
    if (!original_object_ddl_script.empty())
    {
      // reuse the setting schema statement which is the first statement of the alter script
      std::string sql= *statements.begin();
      if ((0 == sql.find("use")) || (0 == sql.find("USE")))
        failback_statements.push_back(sql);
      sql_splitter->splitSqlScript(original_object_ddl_script, failback_statements);
    }
  }

  sql::SqlBatchExec sql_batch_exec;
  sql_batch_exec.stop_on_error(true);
  sql_batch_exec.failback_statements(failback_statements);

  sql_batch_exec.error_cb(on_sql_script_run_error.make_slot());
  sql_batch_exec.batch_exec_progress_cb(on_sql_script_run_progress.make_slot());
  sql_batch_exec.batch_exec_stat_cb(on_sql_script_run_statistics.make_slot());

  if (obj_editor)
  {
    on_sql_script_run_error.connect(obj_editor->on_live_object_change_error);
    on_sql_script_run_progress.connect(obj_editor->on_live_object_change_progress);
    on_sql_script_run_statistics.connect(obj_editor->on_live_object_change_statistics);
  }

  long sql_batch_exec_err_count= 0;
  {
    try
    {
      GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();
      std::auto_ptr<sql::Statement> stmt(_aux_dbc_conn->ref->createStatement());
      sql_batch_exec_err_count= sql_batch_exec(stmt.get(), statements);
    }
    CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Apply object alter script"));
  }

  _history->add_entry(sql_batch_exec.sql_log());

  // refresh object's state only on success, to not lose changes made by user
  if (obj_editor && (0 == sql_batch_exec_err_count))
  {
    db_DatabaseObjectRef db_object= obj_editor->get_dbobject();

    // refresh state of created/altered object in physical overview
    {
      std::string schema_name= db_SchemaRef::can_wrap(db_object) ? std::string() : *db_object->owner()->name();
      db_SchemaRef schema;
      if (!schema_name.empty())
        schema= db_SchemaRef::cast_from(db_object->owner());

      wb::LiveSchemaTree::ObjectType db_object_type;
      if (db_SchemaRef::can_wrap(db_object))
        db_object_type= wb::LiveSchemaTree::Schema;
      else if (db_TableRef::can_wrap(db_object))
        db_object_type= wb::LiveSchemaTree::Table;
      else if (db_ViewRef::can_wrap(db_object))
        db_object_type= wb::LiveSchemaTree::View;
      else if (db_RoutineRef::can_wrap(db_object))
        db_object_type= wb::LiveSchemaTree::Routine;

      refresh_live_object_in_overview(db_object_type, schema_name, db_object->oldName(), db_object->name());
    }

    refresh_live_object_in_editor(obj_editor, false);
  }
}


void Db_sql_editor::create_live_table_stubs(bec::DBObjectEditorBE *table_editor)
{
  db_DatabaseObjectRef db_object= table_editor->get_dbobject();
  if (db_object->customData().has_key("isLiveTableListLoaded"))
    return;

  try
  {
    GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();

    db_CatalogRef catalog= table_editor->get_catalog();
    grt::ListRef<db_Schema> schemata= catalog->schemata();
    db_SchemaRef schema;
    grt::ListRef<db_Table> tables;
    db_TableRef table;

    std::string database_package= *table_editor->get_rdbms()->databaseObjectPackage();
    std::string schema_typename= database_package + ".Schema";
    std::string table_typename= database_package + ".Table";
    grt::GRT *grt= _grtm->get_grt();

    std::string prev_schema_name;

    boost::shared_ptr<sql::ResultSet> rs;
    {
      std::string schema_name= db_SchemaRef::cast_from(db_object->owner())->name();
      std::list<sql::SQLString> table_types;
      table_types.push_back("TABLE");
      rs.reset(_aux_dbc_conn->ref->getMetaData()->getTables("", schema_name, "%", table_types));
    }
    while (rs->next())
    {
      std::string schema_name= rs->getString(2);
      std::string table_name= rs->getString(3);
      if (prev_schema_name != schema_name)
      {
        schema= find_named_object_in_list(schemata, schema_name);
        if (!schema.is_valid())
        {
          schema= grt->create_object<db_Schema>(schema_typename);
          schema->owner(catalog);
          schema->name(schema_name);
          schema->oldName(schema_name);
          schema->modelOnly(1);
          schemata.insert(schema);
        }
        tables= schema->tables();
        prev_schema_name= schema_name;
      }
      table= find_named_object_in_list(tables, table_name);
      if (!table.is_valid())
      {
        table= grt->create_object<db_Table>(table_typename);
        table->owner(schema);
        table->name(table_name);
        table->oldName(table_name);
        table->modelOnly(1);
        table->isStub(1);
        tables.insert(table);
      }
    }

    db_object->customData().set("isLiveTableListLoaded", IntegerRef(1));
  }
  CATCH_ANY_EXCEPTION_AND_DISPATCH(_("Create live table stub"));
}


bool Db_sql_editor::expand_live_table_stub(bec::DBObjectEditorBE *table_editor, const std::string &schema_name, const std::string &obj_name)
{
  db_CatalogRef catalog= table_editor->get_catalog();
  db_TableRef table;
  db_SchemaRef schema= find_named_object_in_list(catalog->schemata(), schema_name);
  if (schema.is_valid())
  {
    table= find_named_object_in_list(schema->tables(), obj_name);
    if (table.is_valid() && table->customData().has_key("isStubExpanded"))
      return true; // stub table has already been expanded
  }

  std::string ddl_script;
  try
  {
    GMutexLock aux_dbc_conn_mutex= ensure_valid_aux_connection();

    std::auto_ptr<sql::ResultSet> rs(_aux_dbc_conn->ref->getMetaData()->getSchemaObjects("", schema_name, "table", true, obj_name));
    if (rs->next())
    {
      ddl_script= strfmt("use `%s`;\n", schema_name.c_str());
      ddl_script+= rs->getString(5);
    }
  }
  catch(const sql::SQLException&) {}

  if (ddl_script.empty())
    return false;

  {
    SqlFacade::Ref sql_facade= SqlFacade::instance_for_rdbms(table_editor->get_rdbms());
    Sql_parser::Ref sql_parser= sql_facade->sqlParser();
    sql_parser->messages_enabled(false);
    grt::DictRef options(_grtm->get_grt());
    {
      std::string sql_mode;
      if (get_session_variable(_aux_dbc_conn, "sql_mode", sql_mode))
        options.gset("sql_mode", sql_mode);
    }
    sql_parser->parse_sql_script(catalog, ddl_script, options);
  }

  // find parsed table
  if (!schema.is_valid())
    schema= find_named_object_in_list(catalog->schemata(), schema_name);
  if (!table.is_valid() && schema.is_valid())
    table= find_named_object_in_list(schema->tables(), obj_name);

  if (table.is_valid() && table != table_editor->get_dbobject())
  {
    table->modelOnly(0);
    table->isStub(1);
    table->customData().set("isStubExpanded", IntegerRef(1));
  }

  return table.is_valid();
}


db_mgmt_RdbmsRef Db_sql_editor::rdbms()
{
  if (_connection.is_valid())
  {
    if (!_connection->driver().is_valid())
      throw std::runtime_error("Connection has invalid driver, check connection parameters.");
    return db_mgmt_RdbmsRef::cast_from(_connection->driver()->owner());
  }
  else
    return db_mgmt_RdbmsRef::cast_from(_grtm->get_grt()->get("/wb/doc/physicalModels/0/rdbms"));
}


std::string Db_sql_editor::caption() const
{
  std::string caption= _("SQL Editor");
  if (*_connection->name().c_str())
    caption+= strfmt(" (%s)", _connection->name().c_str());
  return caption;
}


bool Db_sql_editor::save_snippet(const std::string &text)
{
  if (text.empty()) return false;
  
  std::string name;
  if (!wbsql()->get_wbui()->get_wb()->request_input(_("Enter a Name for the Snippet (optional):"), 0, name))
    return false;

  DbSqlEditorSnippets::get_instance()->add_snippet(name, text);
  do_partial_ui_refresh(RefreshSnippets);
  
  return true;
}

//--------------------------------------------------------------------------------------------------

//mforms::TaskSidebar* Db_sql_editor::get_sidebar()
//{
//  return _side_bar;
//}

//--------------------------------------------------------------------------------------------------

bool Db_sql_editor::is_physical_overview_enabled() const
{
  return _is_physical_overview_enabled;
}


LivePhysicalOverviewBE * Db_sql_editor::live_physical_overview()
{
  return _live_physical_overview;
}


bool Db_sql_editor::activate_live_object(GrtObjectRef object)
{
  std::string obj_name= *object->name();
  std::string owner_name= *object->owner()->name();

  if (db_SchemaRef::can_wrap(object))
    schema_object_activated("activate", LiveSchemaTree::Schema, "", obj_name);
  else if (db_TableRef::can_wrap(object))
    schema_object_activated("activate", LiveSchemaTree::Table, owner_name, obj_name);
  else if (db_ViewRef::can_wrap(object))
    schema_object_activated("activate", LiveSchemaTree::View, owner_name, obj_name);
  else if (db_RoutineRef::can_wrap(object))
    schema_object_activated("activate", LiveSchemaTree::Routine, owner_name, obj_name);
  else
    return false;

  return true;
}


bool Db_sql_editor::create_live_object(GrtObjectRef object_type, std::string owner_name, std::string obj_name)
{
  if (db_SchemaRef::can_wrap(object_type))
    do_create_live_object(LiveSchemaTree::Schema, "", "");
  else if (db_TableRef::can_wrap(object_type))
    do_create_live_object(LiveSchemaTree::Table, owner_name, "");
  else if (db_ViewRef::can_wrap(object_type))
    do_create_live_object(LiveSchemaTree::View, owner_name, "");
  else if (db_RoutineRef::can_wrap(object_type))
    do_create_live_object(LiveSchemaTree::Routine, owner_name, "");
  else
    return false;

  return true;
}


bool Db_sql_editor::drop_live_object(GrtObjectRef object_type, std::string owner_name, std::string obj_name)
{
  if (db_SchemaRef::can_wrap(object_type))
    do_drop_live_object(LiveSchemaTree::Schema, "", obj_name);
  else if (db_TableRef::can_wrap(object_type))
    do_drop_live_object(LiveSchemaTree::Table, owner_name, "");
  else if (db_ViewRef::can_wrap(object_type))
    do_drop_live_object(LiveSchemaTree::View, owner_name, "");
  else if (db_RoutineRef::can_wrap(object_type))
    do_drop_live_object(LiveSchemaTree::Routine, owner_name, "");
  else
    return false;

  return true;
}


void Db_sql_editor::refresh_physical_overview(bool sync)
{
  live_schemata_refresh_task->exec(sync,
    sigc::bind(sigc::mem_fun(this, (grt::StringRef(Db_sql_editor::*)(grt::GRT *, Ptr, bool))&Db_sql_editor::do_refresh_physical_overview),
      weak_ptr_from(this), true));
}


grt::StringRef Db_sql_editor::do_refresh_physical_overview(grt::GRT *grt, Ptr self_ptr, bool hard_refresh)
{
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Db_sql_editor, self_ptr, self, grt::StringRef(""))

  do_refresh_physical_overview(hard_refresh);

  return grt::StringRef("");
}


void Db_sql_editor::do_refresh_physical_overview(bool hard_refresh)
{
  live_schemata_refresh_task->execute_in_main_thread(
    _live_physical_overview->pre_refresh_groups,
    true,
    false);
  
  if (hard_refresh)
  {
    // remove all existing schemata
    std::string database_package= rdbms()->databaseObjectPackage();
    GMutexLock live_physical_overview_model_lock= _live_physical_overview->lock_model();
    workbench_physical_ModelRef model= workbench_physical_ModelRef::cast_from(_live_physical_overview->get_model());
    db_CatalogRef catalog= model->catalog();
    db_SchemaRef schema= _grtm->get_grt()->create_object<db_Schema>(database_package + ".Schema");
    catalog->defaultSchema(schema);
    catalog->schemata().remove_all();
  }

  std::list<std::string> schema_list = fetch_live_schema_list();

  live_schemata_refresh_task->execute_in_main_thread(
    sigc::bind(
      sigc::mem_fun(this, (void(Db_sql_editor::*)(std::list<std::string>, bool))&Db_sql_editor::do_refresh_physical_overview),
      schema_list,
      hard_refresh),
    false,
    false);
}


void Db_sql_editor::do_refresh_physical_overview(std::list<std::string> schema_list, bool hard_refresh)
{
  _schema_tree.refresh_ex(schema_list, hard_refresh);

  _live_physical_overview->refresh_schemata_node();
  _live_physical_overview->send_refresh_schema_list();
  do_partial_ui_refresh(RefreshOverview);

#ifdef _WIN32
  do_ui_refresh();
#else
  do_partial_ui_refresh(RefreshSidebar);
#endif
}

bool Db_sql_editor::can_close()
{  
  return can_close_(true);
}


bool Db_sql_editor::can_close_(bool interactive)
{
  bool check_editors = false;
  
  if (!bec::UIForm::can_close())
    return false;

  if (interactive)
  {
    ConfirmSaveDialog dialog(0, "Close SQL Editor", "The following files/resultsets have unsaved changes.\nDo you want to review these changes before closing?");
    if (check_editors)
    {
      for (int i = 0; i < sql_editor_count(); i++)
      {
        if (sql_editor_dirty(i))
        {
          std::string n= sql_editor_path(i);
          if (!n.empty())
            n= std::string(g_basename(n.c_str())).append(" - ").append(n);
          else
            n= "Unsaved SQL Query";
          dialog.add_item("Script Buffers", n);
        }
      }
    }
    
    Recordsets recordsets;
    {
      GMutexLock recordsets_mutex(_recordsets_mutex);
      recordsets= _recordsets;
    }
    BOOST_FOREACH (Recordset::Ref rs, recordsets)
      if (!rs->can_close(false))
        dialog.add_item("Resultset", rs->caption());

    bool review= false;
    if (dialog.change_count() > 1)
    {
      switch (dialog.run()) 
      {
        case ConfirmSaveDialog::ReviewChanges:
          review= true;
          break;

        case ConfirmSaveDialog::DiscardChanges:
          review= false;
          break;

        case ConfirmSaveDialog::Cancel:
          return false;
      }
    }
    else if (dialog.change_count() == 1)
      review= true;
     
    // review changes 1 by 1
    if (review && check_editors)
    {
      for (int i = 0; i < sql_editor_count(); i++)
      {
        if (sql_editor_dirty(i))
        {
          if (!sql_editor_will_close(i))
            return false;
        }
      }
      
      BOOST_FOREACH (Recordset::Ref rs, recordsets)
        if (!rs->can_close(true))
          return false;      
    }
  }
  else
  {
    if (check_editors)
    {
      for (int i = 0; i < sql_editor_count(); i++)
      {
        if (sql_editor_dirty(i))
          return false;
      }
    }
    
    Recordsets recordsets;
    {
      GMutexLock recordsets_mutex(_recordsets_mutex);
      recordsets= _recordsets;
    }
    BOOST_FOREACH (Recordset::Ref rs, recordsets)
      if (!rs->can_close(false))
        return false;
  }

  return true;
}



bec::MenuItemList Db_sql_editor::get_live_catalog_menu_items(const std::string &schema, 
                                                               const std::string &object, 
                                                               const std::string &type)
{
  grt::ListRef<GrtObject> list(_context_ui->get_wb()->get_grt());
  std::list<std::string> groups;
  
  db_query_LiveDBObjectRef dbobject(list.get_grt());
  dbobject->owner(_wbsql->get_grt_editor_object(this));
  dbobject->schemaName(schema);
  dbobject->type(type);
  dbobject->name(object);
  if (type == "schema")
    dbobject->name(schema);
  list.insert(dbobject);
  
  bec::ArgumentPool argument_pool;
  _context_ui->get_wb()->update_plugin_arguments_pool(argument_pool);
  
  argument_pool.add_entries_for_object("", dbobject, "");
  argument_pool.add_entries_for_object(type, dbobject, "");
  
  groups.push_back("Menu/SQL/Catalog");
  
  return grt_manager()->get_plugin_context_menu_items(groups, argument_pool);
}


bool Db_sql_editor::call_live_catalog_menu_item(const std::string &item,
                                                const std::string &schema, const std::string &object, 
                                                const std::string &type)
{  
  db_query_LiveDBObjectRef dbobject(_context_ui->get_wb()->get_grt());
  dbobject->owner(_wbsql->get_grt_editor_object(this));
  dbobject->schemaName(schema);
  dbobject->type(type);
  dbobject->name(object);
  if (type == "schema")
    dbobject->name(schema);
  
  bec::ArgumentPool argument_pool;
  _context_ui->get_wb()->update_plugin_arguments_pool(argument_pool);
  
  argument_pool.add_entries_for_object("", dbobject, "");
  argument_pool.add_entries_for_object(type, dbobject, "");

  _context_ui->get_command_ui()->activate_command(item, argument_pool);
  return true;
}


bec::MenuItemList Db_sql_editor::recordset_popup_menu_items(const std::vector<int> &rows, int column, Recordset::Ptr rs_ptr)
{
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs, bec::MenuItemList())
  if (rs)
  {
    bec::ArgumentPool argpool;
    
    if (!rows.empty())
    {
      grt::IntegerListRef row_list(grt_manager()->get_grt());
      for (std::vector<int>::const_iterator iter= rows.begin(); iter != rows.end(); ++iter)
        row_list.insert(grt::IntegerRef(*iter));
      argpool.add_simple_value("selectedRows", row_list);
      argpool.add_simple_value("clickedRow", grt::IntegerRef(rows[0]));
    }
    argpool.add_simple_value("clickedColumn", grt::IntegerRef(column));

    _context_ui->get_wb()->update_plugin_arguments_pool(argpool);

    db_query_EditorRef editor(_wbsql->get_grt_editor_object(this));
    
    for (size_t c= editor->resultsets().count(), i= 0; i < c; i++)
    {
      db_query_ResultsetRef rset(editor->resultsets()[i]);
      if (rset->get_data()->recordset.get() == rs)
      {
        // add rset to the pool, overriding the one added by update_args
        argpool.add_entries_for_object("", rset, "db.query.Resultset");
        break;
      }
    }

    std::list<std::string> groups;
    groups.push_back("Menu/SQL/Resultset");

    return grt_manager()->get_plugin_context_menu_items(groups, argpool);    
  }
  
  return bec::MenuItemList();
}


bool Db_sql_editor::call_recordset_popup_menu_item(const std::string &item, const std::vector<int> &rows,
                                                   int column, Recordset::Ptr rs_ptr)
{
  RETVAL_IF_FAIL_TO_RETAIN_WEAK_PTR (Recordset, rs_ptr, rs, false)
  
  if (rs)
  {
    bec::ArgumentPool argpool;
    
    if (!rows.empty())
    {
      grt::IntegerListRef row_list(grt_manager()->get_grt());
      for (std::vector<int>::const_iterator iter= rows.begin(); iter != rows.end(); ++iter)
        row_list.insert(grt::IntegerRef(*iter));
      argpool["app.PluginInputDefinition:selectedRows"]= row_list;
      argpool["app.PluginInputDefinition:clickedRow"]= grt::IntegerRef(rows[0]);
    }
    argpool["app.PluginInputDefinition:clickedColumn"]= grt::IntegerRef(column);
    
    _context_ui->get_wb()->update_plugin_arguments_pool(argpool);
    
    db_query_EditorRef editor(_wbsql->get_grt_editor_object(this));
    
    for (size_t c= editor->resultsets().count(), i= 0; i < c; i++)
    {
      db_query_ResultsetRef rset(editor->resultsets()[i]);
      if (rset->get_data()->recordset.get() == rs)
      {
        // add rset to the pool, overriding the one added by update_args
        argpool.add_entries_for_object("", rset, "db.query.Resultset");
        break;
      }
    }
    
    return _context_ui->get_command_ui()->activate_command(item, argpool);
  }
  
  return false;
}


void Db_sql_editor::update_menu_and_toolbar()
{
  if (!_grtm->in_main_thread())
  {
    exec_sql_task->execute_in_main_thread(sigc::mem_fun(this, &Db_sql_editor::update_menu_and_toolbar),
                                          false,
                                          false);
    return;
  }
  
  bool running = is_running_query();
  
  if (_menu)
  {
    _menu->set_item_enabled("query.cancel", running);
    _menu->set_item_enabled("query.execute", !running);
    _menu->set_item_enabled("query.reconnect", !running);
    _menu->set_item_enabled("wb.sqlide.executeToTextOutput", !running);
    _menu->set_item_enabled("query.execute_current_statement", !running);
    _menu->set_item_enabled("query.explain", !running);
    _menu->set_item_enabled("query.explain_current_statement", !running);

    _menu->set_item_enabled("query.commit", !running && !auto_commit());
    _menu->set_item_enabled("query.rollback", !running && !auto_commit());    
    mforms::MenuItem *item = _menu->find_item("query.autocommit");
    if (item)
    {
      item->set_enabled(!running);
      item->set_checked(auto_commit());    
    }
  }

  if (_toolbar)
  {
    _toolbar->set_item_enabled("query.cancel", running);
    _toolbar->set_item_enabled("query.execute", !running);
    _toolbar->set_item_enabled("query.reconnect", !running);
    _toolbar->set_item_enabled("wb.sqlide.executeToTextOutput", !running);
    _toolbar->set_item_enabled("query.execute_current_statement", !running);
    _toolbar->set_item_enabled("query.explain", !running);
    _toolbar->set_item_enabled("query.explain_current_statement", !running);
    
    _toolbar->set_item_enabled("query.commit", !running && !auto_commit());
    _toolbar->set_item_enabled("query.rollback", !running && !auto_commit());
    mforms::ToolBarItem *item = _toolbar->find_item("query.autocommit");
    if (item)
    {
      item->set_enabled(!running);
      item->set_checked(auto_commit());
    }
    
    item = _toolbar->find_item("query.recordsets_are_pinned_by_default");
    if (item)
      item->set_checked(_recordsets_are_pinned_by_default);
  }
}


void Db_sql_editor::main_form_changed(bec::UIForm *form)
{
  if (form == this)
  {
    // revalidate menu items
    update_menu_and_toolbar();
  }
}

//--------------------------------------------------------------------------------------------------

void Db_sql_editor::setup_sidebar()
{
  _side_bar = mforms::TaskSidebar::create();
  _side_bar->set_schema_model(&_schema_tree);
  _side_bar->on_section_command().connect(sigc::mem_fun(this, &Db_sql_editor::sidebar_command));

  std::string section = _("ACTIONS");
  std::string icon_suffix;
#ifdef _WIN32
  icon_suffix = "_win.png";
#else
  icon_suffix = ".png";
#endif
  _side_bar->add_section(section);
  _side_bar->add_section_entry(section, _("Execute SQL File"), "sql_run_script" + icon_suffix, "run_script", true);
  _side_bar->add_section_entry(section, _("Add Schema"), "sql_add_schema" + icon_suffix, "add_schema", true);
  //_side_bar->add_section_entry(section, _("Copy Schema"), "sql_copy_schema" + icon_suffix, "copy_schema");
  _side_bar->add_section_entry(section, _("Add Table"), "sql_add_table" + icon_suffix, "add_table", true);
  _side_bar->add_section_entry(section, _("Add View"), "sql_add_view" + icon_suffix, "add_view", true);
  _side_bar->add_section_entry(section, _("Add Routine"), "sql_add_routine" + icon_suffix, "add_routine", true);
  
  _side_bar->get_schema_tree()->signal_changed().connect(sigc::mem_fun(this, &Db_sql_editor::schema_row_selected));
  
  mforms::Panel* panel;
#ifdef _WIN32
  _side_splitter.set_back_color("#283752");
  panel = mforms::manage(new mforms::Panel(mforms::StyledHeaderPanel));
  panel->set_title("Object Browser");
  panel->add(_side_bar);
  _side_splitter.add(panel, 200);
#else
  _side_splitter.add(_side_bar, 200);
#endif
  panel = mforms::manage(new mforms::Panel(mforms::StyledHeaderPanel));
  panel->set_title("Object Information");
  _info_box.set_size(-1, 200);
  _info_box.set_monospaced(true);
  panel->add(&_info_box);

  _side_splitter.add(panel, 40);

  _side_splitter.set_position(500);

}

//--------------------------------------------------------------------------------------------------

void Db_sql_editor::schema_row_selected()
{
  bec::NodeId node;
  std::string details;
  
  if (!_side_bar->get_schema_tree()->get_selected_node(node))
    details= _connection_info;
  else
    details= get_schema_tree()->get_field_description(node, 0);
  
  _info_box.set_value(details);
}

//--------------------------------------------------------------------------------------------------

void Db_sql_editor::sidebar_command(const std::string& command)
{
  if (command == "add_schema")
    _schema_tree.create_live_object_signal.emit(LiveSchemaTree::Schema, "", "");
  else if (command == "add_table")
    _schema_tree.create_live_object_signal.emit(LiveSchemaTree::Table, "", "");
  else if (command == "add_view")
    _schema_tree.create_live_object_signal.emit(LiveSchemaTree::View, "", "");
  else if (command == "add_routine")
    _schema_tree.create_live_object_signal.emit(LiveSchemaTree::Routine, "", "");
  else if (command == "run_script")
  {
    std::string path = _wbsql->get_wbui()->get_wb()->show_file_dialog("open", _("Execute SQL Script"),
      "SQL Files (*.sql)|*.sql");
    open_sql_script_file(path, true, true);
  }
  else if (command == "copy_schema")
    {} // TODO: implement copy_schema
};

//--------------------------------------------------------------------------------------------------

